import importlib

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal, assert_series_equal

import bofire.data_models.molfeatures.names as names
from bofire.data_models.enum import CategoricalEncodingEnum
from bofire.data_models.features.molecular import (
    CategoricalMolecularInput,
    MolecularInput,
)
from bofire.data_models.molfeatures.api import (
    Fingerprints,
    FingerprintsFragments,
    Fragments,
    MordredDescriptors,
)

RDKIT_AVAILABLE = importlib.util.find_spec("rdkit") is not None

smiles = [
    "CC(=O)Oc1ccccc1C(=O)O",
    "c1ccccc1",
    "[CH3][CH2][OH]",
    "N[C@](C)(F)C(=O)O",
]
VALID_SMILES = pd.Series(smiles)
VALID_SMILES.name = "molecule"
INVALID_SMILES = pd.Series(["CC(=O)Oc1ccccc1C(=O)O", "c1ccccc1", "abcd"])


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_molecular_input_validate_experimental():
    m = MolecularInput(key="molecule")
    vals = m.validate_experimental(VALID_SMILES)
    assert_series_equal(vals, VALID_SMILES)
    with pytest.raises(ValueError):
        m.validate_experimental(INVALID_SMILES)


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_molecular_input_validate_candidental():
    m = MolecularInput(key="molecule")
    vals = m.validate_candidental(VALID_SMILES)
    assert_series_equal(vals, VALID_SMILES)
    with pytest.raises(ValueError):
        m.validate_candidental(INVALID_SMILES)


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_molecular_input_fixed():
    m = MolecularInput(key="molecule")
    assert m.fixed_value() is None
    assert m.is_fixed() is False


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
@pytest.mark.parametrize(
    "molfeatures, expected",
    [
        (Fingerprints(), [f"fingerprint_{i}" for i in range(2048)]),
        (Fingerprints(n_bits=32), [f"fingerprint_{i}" for i in range(32)]),
        (
            Fragments(),
            names.fragments,
        ),
        (
            Fragments(fragments=["fr_unbrch_alkane", "fr_thiocyan"]),
            ["fr_unbrch_alkane", "fr_thiocyan"],
        ),
        (
            FingerprintsFragments(),
            [f"fingerprint_{i}" for i in range(2048)] + names.fragments,
        ),
        (
            FingerprintsFragments(
                n_bits=32, fragments=["fr_unbrch_alkane", "fr_thiocyan"]
            ),
            [f"fingerprint_{i}" for i in range(32)]
            + ["fr_unbrch_alkane", "fr_thiocyan"],
        ),
        (MordredDescriptors(descriptors=["NssCH2", "ATSC2d"]), ["NssCH2", "ATSC2d"]),
    ],
)
def test_molfeatures_get_descriptor_names(molfeatures, expected):
    assert molfeatures.get_descriptor_names() == expected


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
@pytest.mark.parametrize(
    "expected, transform_type",
    [
        (
            (
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    0.0,
                ],
            ),
            Fingerprints(n_bits=32),
        ),
        (
            (
                [0.0, 0.0, 0.0, 0.0],
                [1.0, 1.0, 2.0, 1.0],
            ),
            Fragments(fragments=["fr_COO", "fr_COO2", "fr_C_O", "fr_C_O_noCOO"]),
        ),
        (
            (
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    0.0,
                    1.0,
                    1.0,
                    2.0,
                    1.0,
                ],
            ),
            FingerprintsFragments(
                n_bits=32, fragments=["fr_COO", "fr_COO2", "fr_C_O", "fr_C_O_noCOO"]
            ),
        ),
        (
            ([-8.34319526627219, 0.0], [0.5963718820861676, 1.0]),
            MordredDescriptors(descriptors=["NssCH2", "ATSC2d"]),
        ),
    ],
)
def test_molecular_descriptor_feature_get_bounds(expected, transform_type):
    input_feature = MolecularInput(key="molecule")
    lower, upper = input_feature.get_bounds(
        transform_type=transform_type,
        values=VALID_SMILES,
    )
    assert np.allclose(lower, expected[0])
    assert np.allclose(upper, expected[1])


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
@pytest.mark.parametrize(
    "key, transform_type, values",
    [
        (
            "molecule_2_two",
            Fingerprints(n_bits=32),
            {
                "molecule_2_two_fingerprint_0": {0: 1.0, 1: 1.0, 2: 0.0, 3: 0.0},
                "molecule_2_two_fingerprint_1": {0: 1.0, 1: 0.0, 2: 1.0, 3: 1.0},
                "molecule_2_two_fingerprint_2": {0: 1.0, 1: 0.0, 2: 1.0, 3: 0.0},
                "molecule_2_two_fingerprint_3": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_2_two_fingerprint_4": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_2_two_fingerprint_5": {0: 1.0, 1: 1.0, 2: 0.0, 3: 1.0},
                "molecule_2_two_fingerprint_6": {0: 0.0, 1: 0.0, 2: 1.0, 3: 0.0},
                "molecule_2_two_fingerprint_7": {0: 1.0, 1: 0.0, 2: 1.0, 3: 1.0},
                "molecule_2_two_fingerprint_8": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_2_two_fingerprint_9": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_2_two_fingerprint_10": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_2_two_fingerprint_11": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_2_two_fingerprint_12": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_2_two_fingerprint_13": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_2_two_fingerprint_14": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_2_two_fingerprint_15": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_2_two_fingerprint_16": {0: 1.0, 1: 1.0, 2: 1.0, 3: 0.0},
                "molecule_2_two_fingerprint_17": {0: 1.0, 1: 1.0, 2: 0.0, 3: 0.0},
                "molecule_2_two_fingerprint_18": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_2_two_fingerprint_19": {0: 0.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_2_two_fingerprint_20": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_2_two_fingerprint_21": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_2_two_fingerprint_22": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_2_two_fingerprint_23": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_2_two_fingerprint_24": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_2_two_fingerprint_25": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_2_two_fingerprint_26": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_2_two_fingerprint_27": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_2_two_fingerprint_28": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_2_two_fingerprint_29": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_2_two_fingerprint_30": {0: 0.0, 1: 0.0, 2: 1.0, 3: 0.0},
                "molecule_2_two_fingerprint_31": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
            },
        ),
        (
            "molecule_",
            Fragments(fragments=["fr_unbrch_alkane", "fr_thiocyan"]),
            {
                "molecule__fr_unbrch_alkane": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule__fr_thiocyan": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
            },
        ),
        (
            "molecule",
            FingerprintsFragments(
                n_bits=32, fragments=["fr_unbrch_alkane", "fr_thiocyan"]
            ),
            {
                "molecule_fingerprint_0": {0: 1.0, 1: 1.0, 2: 0.0, 3: 0.0},
                "molecule_fingerprint_1": {0: 1.0, 1: 0.0, 2: 1.0, 3: 1.0},
                "molecule_fingerprint_2": {0: 1.0, 1: 0.0, 2: 1.0, 3: 0.0},
                "molecule_fingerprint_3": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_fingerprint_4": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_fingerprint_5": {0: 1.0, 1: 1.0, 2: 0.0, 3: 1.0},
                "molecule_fingerprint_6": {0: 0.0, 1: 0.0, 2: 1.0, 3: 0.0},
                "molecule_fingerprint_7": {0: 1.0, 1: 0.0, 2: 1.0, 3: 1.0},
                "molecule_fingerprint_8": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_fingerprint_9": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_fingerprint_10": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_fingerprint_11": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_fingerprint_12": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_fingerprint_13": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_fingerprint_14": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_fingerprint_15": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_fingerprint_16": {0: 1.0, 1: 1.0, 2: 1.0, 3: 0.0},
                "molecule_fingerprint_17": {0: 1.0, 1: 1.0, 2: 0.0, 3: 0.0},
                "molecule_fingerprint_18": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_fingerprint_19": {0: 0.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_fingerprint_20": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_fingerprint_21": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_fingerprint_22": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_fingerprint_23": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_fingerprint_24": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_fingerprint_25": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_fingerprint_26": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_fingerprint_27": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_fingerprint_28": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_fingerprint_29": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
                "molecule_fingerprint_30": {0: 0.0, 1: 0.0, 2: 1.0, 3: 0.0},
                "molecule_fingerprint_31": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_fr_unbrch_alkane": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                "molecule_fr_thiocyan": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
            },
        ),
        (
            "_mo_le_cule",
            MordredDescriptors(descriptors=["NssCH2", "ATSC2d"]),
            {
                "_mo_le_cule_NssCH2": {
                    0: 0.5963718820861676,
                    1: -1.5,
                    2: -0.28395061728395066,
                    3: -8.34319526627219,
                },
                "_mo_le_cule_ATSC2d": {0: 0.0, 1: 0.0, 2: 1.0, 3: 0.0},
            },
        ),
    ],
)
def test_molecular_input_to_descriptor_encoding(key, transform_type, values):
    input_feature = MolecularInput(key=key)

    encoded = input_feature.to_descriptor_encoding(transform_type, VALID_SMILES)
    assert len(encoded.columns) == len(transform_type.get_descriptor_names())
    assert len(encoded) == len(smiles)
    assert_frame_equal(encoded, pd.DataFrame.from_dict(values))


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_molfeatures_type_get_descriptor_values_fingerprints():
    values = {
        "fingerprint_0": {0: 1.0, 1: 1.0, 2: 0.0, 3: 0.0},
        "fingerprint_1": {0: 1.0, 1: 0.0, 2: 1.0, 3: 1.0},
        "fingerprint_2": {0: 1.0, 1: 0.0, 2: 1.0, 3: 0.0},
        "fingerprint_3": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_4": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_5": {0: 1.0, 1: 1.0, 2: 0.0, 3: 1.0},
        "fingerprint_6": {0: 0.0, 1: 0.0, 2: 1.0, 3: 0.0},
        "fingerprint_7": {0: 1.0, 1: 0.0, 2: 1.0, 3: 1.0},
        "fingerprint_8": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_9": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_10": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_11": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_12": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_13": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_14": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_15": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_16": {0: 1.0, 1: 1.0, 2: 1.0, 3: 0.0},
        "fingerprint_17": {0: 1.0, 1: 1.0, 2: 0.0, 3: 0.0},
        "fingerprint_18": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_19": {0: 0.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_20": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_21": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_22": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_23": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_24": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_25": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_26": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_27": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_28": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_29": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_30": {0: 0.0, 1: 0.0, 2: 1.0, 3: 0.0},
        "fingerprint_31": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
    }

    molfeature = Fingerprints(n_bits=32)
    generated = molfeature.get_descriptor_values(VALID_SMILES)
    assert_frame_equal(generated, pd.DataFrame.from_dict(values))


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_molfeatures_type_get_descriptor_values_fragments():
    values = {
        "fr_unbrch_alkane": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fr_thiocyan": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
    }

    molfeature = Fragments(fragments=["fr_unbrch_alkane", "fr_thiocyan"])
    generated = molfeature.get_descriptor_values(VALID_SMILES)
    assert_frame_equal(generated, pd.DataFrame.from_dict(values))


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
@pytest.mark.parametrize(
    "fragment_list",
    [
        (["fr_unbrch_alkane','fr_unbrch_alkane', 'fr_thiocyan"]),
        (["frag','fr_unbrch_alkane', 'fr_thiocyan"]),
    ],
)
def test_molfeatures_type_fragments_invalid(fragment_list):
    with pytest.raises(ValueError):
        FingerprintsFragments(fragments=fragment_list)


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_molfeatures_type_get_descriptor_values_fingerprintsfragments():
    values = {
        "fingerprint_0": {0: 1.0, 1: 1.0, 2: 0.0, 3: 0.0},
        "fingerprint_1": {0: 1.0, 1: 0.0, 2: 1.0, 3: 1.0},
        "fingerprint_2": {0: 1.0, 1: 0.0, 2: 1.0, 3: 0.0},
        "fingerprint_3": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_4": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_5": {0: 1.0, 1: 1.0, 2: 0.0, 3: 1.0},
        "fingerprint_6": {0: 0.0, 1: 0.0, 2: 1.0, 3: 0.0},
        "fingerprint_7": {0: 1.0, 1: 0.0, 2: 1.0, 3: 1.0},
        "fingerprint_8": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_9": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_10": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_11": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_12": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_13": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_14": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_15": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_16": {0: 1.0, 1: 1.0, 2: 1.0, 3: 0.0},
        "fingerprint_17": {0: 1.0, 1: 1.0, 2: 0.0, 3: 0.0},
        "fingerprint_18": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_19": {0: 0.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_20": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_21": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_22": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_23": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_24": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_25": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_26": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_27": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_28": {0: 1.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fingerprint_29": {0: 1.0, 1: 0.0, 2: 0.0, 3: 1.0},
        "fingerprint_30": {0: 0.0, 1: 0.0, 2: 1.0, 3: 0.0},
        "fingerprint_31": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fr_unbrch_alkane": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
        "fr_thiocyan": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
    }

    molfeature = FingerprintsFragments(
        n_bits=32, fragments=["fr_unbrch_alkane", "fr_thiocyan"]
    )
    generated = molfeature.get_descriptor_values(VALID_SMILES)
    assert_frame_equal(generated, pd.DataFrame.from_dict(values))


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
@pytest.mark.parametrize(
    "fragment_list",
    [
        (["fr_unbrch_alkane','fr_unbrch_alkane', 'fr_thiocyan"]),
        (["frag','fr_unbrch_alkane', 'fr_thiocyan"]),
    ],
)
def test_molfeatures_type_fingerprintsfragments_invalid(fragment_list):
    with pytest.raises(ValueError):
        FingerprintsFragments(fragments=fragment_list)


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_molfeatures_type_get_descriptor_values_mordreddescriptors():
    values = {
        "NssCH2": {
            0: 0.5963718820861676,
            1: -1.5,
            2: -0.28395061728395066,
            3: -8.34319526627219,
        },
        "ATSC2d": {0: 0.0, 1: 0.0, 2: 1.0, 3: 0.0},
    }

    molfeature = MordredDescriptors(descriptors=["NssCH2", "ATSC2d"])
    generated = molfeature.get_descriptor_values(VALID_SMILES)
    assert_frame_equal(generated, pd.DataFrame.from_dict(values))


@pytest.mark.parametrize(
    "mordred_list",
    [
        (["NssCH2", "NssCH2", "ATSC2d"]),
        (["desc", "NssCH2", "ATSC2d"]),
    ],
)
@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_molfeatures_type_mordreddescriptors_invalid(mordred_list):
    with pytest.raises(ValueError):
        MordredDescriptors(descriptors=mordred_list)


### tests for CategoricalMolecularInput
@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_categorical_molecular_input_invalid_smiles():
    with pytest.raises(ValueError, match="abcd is not a valid smiles string."):
        CategoricalMolecularInput(
            key="a", categories=["CC(=O)Oc1ccccc1C(=O)O", "c1ccccc1", "abcd"]
        )


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_categorical_molecular_input_valid_smiles():
    CategoricalMolecularInput(key="a", categories=VALID_SMILES.tolist())


@pytest.mark.parametrize(
    "key",
    [
        ("molecule_2_two"),
        ("molecule_"),
        ("molecule"),
        ("_mo_le_cule"),
    ],
)
@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_categorical_molecular_input_from_descriptor_encoding(key):
    feat = CategoricalMolecularInput(key=key, categories=VALID_SMILES.to_list())
    values = pd.Series(data=["c1ccccc1", "[CH3][CH2][OH]"], name=key)
    for transform_type in [
        Fingerprints(),
        FingerprintsFragments(),
        Fragments(),
        MordredDescriptors(descriptors=["NssCH2", "ATSC2d"]),
    ]:
        encoded = feat.to_descriptor_encoding(transform_type, values=values)
        decoded = feat.from_descriptor_encoding(transform_type, values=encoded)
        assert np.all(decoded == values)


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_categorical_molecular_input_get_bounds():
    # first test with onehot
    feat = CategoricalMolecularInput(
        key="a", categories=VALID_SMILES.to_list(), allowed=[True, True, True, True]
    )
    lower, upper = feat.get_bounds(transform_type=CategoricalEncodingEnum.ONE_HOT)
    assert lower == [0 for _ in range(len(feat.categories))]
    assert upper == [1 for _ in range(len(feat.categories))]
    # now test it with descriptors,
    feat = CategoricalMolecularInput(
        key="a", categories=VALID_SMILES.to_list(), allowed=[True, True, False, False]
    )
    lower, upper = feat.get_bounds(
        transform_type=MordredDescriptors(
            descriptors=[
                "nAromAtom",
                "nAromBond",
            ]
        )
    )
    assert lower == [6.0, 6.0]
    assert upper == [6.0, 6.0]

    lower, upper = feat.get_bounds(
        transform_type=MordredDescriptors(
            descriptors=[
                "nAromAtom",
                "nAromBond",
            ]
        ),
        values=VALID_SMILES,
    )
    assert lower == [0.0, 0.0]
    assert upper == [6.0, 6.0]



from copy import deepcopy
from typing import List

from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.domain.domain import Domain
from bofire.data_models.features.api import Input


def get_subdomain(
    domain: Domain,
    feature_keys: List,
) -> Domain:
    """removes all features not defined as argument creating a subdomain of the provided domain

    Args:
        domain (Domain): the original domain wherefrom a subdomain should be created
        feature_keys (List): List of features that shall be included in the subdomain

    Raises:
        Assert: when in total less than 2 features are provided
        ValueError: when a provided feature key is not present in the provided domain
        Assert: when no output feature is provided
        Assert: when no input feature is provided
        ValueError: _description_

    Returns:
        Domain: A new domain containing only parts of the original domain
    """
    assert len(feature_keys) >= 2, "At least two features have to be provided."
    outputs = []
    inputs = []
    for key in feature_keys:
        try:
            feat = domain.get_feature(key)
        except KeyError:
            raise ValueError(f"Feature {key} not present in domain.")
        if isinstance(feat, Input):
            inputs.append(feat)
        else:
            outputs.append(feat)
    assert len(outputs) > 0, "At least one output feature has to be provided."
    assert len(inputs) > 0, "At least one input feature has to be provided."
    inputs = Inputs(features=inputs)
    outputs = Outputs(features=outputs)
    # loop over constraints and make sure that all features used in constraints are in the input_feature_keys
    for c in domain.constraints:
        for key in c.features:  # type: ignore
            if key not in inputs.get_keys():
                raise ValueError(
                    f"Removed input feature {key} is used in a constraint."
                )
    subdomain = deepcopy(domain)
    subdomain.inputs = inputs
    subdomain.outputs = outputs
    return subdomain



from copy import deepcopy
from typing import List, Tuple, cast

import numpy as np
import pandas as pd

from bofire.data_models.constraints.api import (
    AnyConstraint,
    Constraint,
    LinearConstraint,
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NChooseKConstraint,
)
from bofire.data_models.domain.api import Constraints, Domain, Inputs
from bofire.data_models.features.api import ContinuousInput, Input

### this module is based on the original implementation in basf/opti.


class AffineTransform:
    """Class to switch back and forth from the reduced to the original domain."""

    def __init__(self, equalities: List[Tuple[str, List[str], List[float]]]):
        """Initializes a `AffineTransformation` object.

        Args:
            equalities (List[Tuple[str,List[str],List[float]]]): List of equalities. Every equality
                is defined as a tuple, in which the first entry is the key of the reduced feature, the second
                one is a list of feature keys that can be used to compute the feature and the third list of floats
                are the corresponding coefficients.
        """
        self.equalities = equalities

    def augment_data(self, data: pd.DataFrame) -> pd.DataFrame:
        """Restore the eliminated features in a dataframe

        Args:
            data (pd.DataFrame): Dataframe that should be restored.

        Returns:
            pd.DataFrame: Restored dataframe
        """
        if len(self.equalities) == 0:
            return data
        data = data.copy()
        for name_lhs, names_rhs, coeffs in self.equalities:
            data[name_lhs] = coeffs[-1]
            for i, name in enumerate(names_rhs):
                data[name_lhs] += coeffs[i] * data[name]
        return data

    def drop_data(self, data: pd.DataFrame) -> pd.DataFrame:
        """Drop eliminated features from a dataframe.

        Args:
            data (pd.DataFrame): Dataframe with features to be dropped.

        Returns:
            pd.DataFrame: Reduced dataframe.
        """
        if len(self.equalities) == 0:
            return data
        drop = []
        for name_lhs, _, _ in self.equalities:
            if name_lhs in data.columns:
                drop.append(name_lhs)
        return data.drop(columns=drop)


def reduce_domain(domain: Domain) -> Tuple[Domain, AffineTransform]:
    """Reduce a domain with linear equality constraints to a subdomain where linear equality constraints are eliminated.

    Args:
        domain (Domain): Domain to be reduced.

    Returns:
        Tuple[Domain, AffineTransform]: reduced domain and the according transformation to switch between the
            reduced and orginal domain.
    """
    # check if the domain can be reduced
    if not check_domain_for_reduction(domain):
        return domain, AffineTransform([])

    # find linear equality constraints
    linear_equalities = domain.constraints.get(LinearEqualityConstraint)
    other_constraints = domain.constraints.get(
        Constraint, excludes=[LinearEqualityConstraint]
    )

    # only consider continuous inputs
    continuous_inputs = [
        cast(ContinuousInput, f) for f in domain.get_features(ContinuousInput)
    ]
    other_inputs = domain.inputs.get(Input, excludes=[ContinuousInput])

    # assemble Matrix A from equality constraints
    N = len(linear_equalities)
    M = len(continuous_inputs) + 1
    names = np.concatenate(([feat.key for feat in continuous_inputs], ["rhs"]))

    A_aug = pd.DataFrame(data=np.zeros(shape=(N, M)), columns=names)

    for i in range(len(linear_equalities)):
        c = linear_equalities[i]
        assert isinstance(c, LinearEqualityConstraint)
        A_aug.loc[i, c.features] = c.coefficients  # type: ignore
        A_aug.loc[i, "rhs"] = c.rhs
    A_aug = A_aug.values

    # catch special cases
    check_existence_of_solution(A_aug)

    # bring A_aug to reduced row-echelon form
    A_aug_rref, pivots = rref(A_aug)
    pivots = np.array(pivots)
    A_aug_rref = np.array(A_aug_rref).astype(np.float64)

    # formulate box bounds as linear inequality constraints in matrix form
    B = np.zeros(shape=(2 * (M - 1), M))
    B[: M - 1, : M - 1] = np.eye(M - 1)
    B[M - 1 :, : M - 1] = -np.eye(M - 1)

    B[: M - 1, -1] = np.array([feat.upper_bound for feat in continuous_inputs])
    B[M - 1 :, -1] = -1.0 * np.array([feat.lower_bound for feat in continuous_inputs])

    # eliminate columns with pivot element
    for i in range(len(pivots)):
        p = pivots[i]
        B[p, :] -= A_aug_rref[i, :]
        B[p + M - 1, :] += A_aug_rref[i, :]

    # build up reduced domain
    _domain = Domain.construct(
        # _fields_set = {"inputs", "outputs", "constraints"}
        inputs=deepcopy(other_inputs),
        outputs=deepcopy(domain.outputs),
        constraints=deepcopy(other_constraints),
    )
    new_inputs = [
        deepcopy(feat) for i, feat in enumerate(continuous_inputs) if i not in pivots
    ]
    all_inputs = _domain.inputs + new_inputs
    assert isinstance(all_inputs, Inputs)
    _domain.inputs = all_inputs

    constraints: List[AnyConstraint] = []
    for i in pivots:
        # reduce equation system of upper bounds
        ind = np.where(B[i, :-1] != 0)[0]
        if len(ind) > 0 and B[i, -1] < np.inf:
            if len(list(names[ind])) > 1:
                c = LinearInequalityConstraint.from_greater_equal(
                    features=list(names[ind]),
                    coefficients=(-1.0 * B[i, ind]).tolist(),
                    rhs=B[i, -1] * -1.0,
                )
                constraints.append(c)
            else:
                key = names[ind][0]
                feat = cast(ContinuousInput, _domain.get_feature(key))
                adjust_boundary(feat, (-1.0 * B[i, ind])[0], B[i, -1] * -1.0)
        else:
            if B[i, -1] < -1e-16:
                raise Exception("There is no solution that fulfills the constraints.")

        # reduce equation system of lower bounds
        ind = np.where(B[i + M - 1, :-1] != 0)[0]
        if len(ind) > 0 and B[i + M - 1, -1] < np.inf:
            if len(list(names[ind])) > 1:
                c = LinearInequalityConstraint.from_greater_equal(
                    features=list(names[ind]),
                    coefficients=(-1.0 * B[i + M - 1, ind]).tolist(),
                    rhs=B[i + M - 1, -1] * -1.0,
                )
                constraints.append(c)
            else:
                key = names[ind][0]
                feat = cast(ContinuousInput, _domain.get_feature(key))
                adjust_boundary(
                    feat,
                    (-1.0 * B[i + M - 1, ind])[0],
                    B[i + M - 1, -1] * -1.0,
                )
        else:
            if B[i + M - 1, -1] < -1e-16:
                raise Exception("There is no solution that fulfills the constraints.")

    if len(constraints) > 0:
        _domain._set_constraints_unvalidated(_domain.constraints + constraints)

    # assemble equalities
    _equalities = []
    for i in range(len(pivots)):
        name_lhs = names[pivots[i]]
        names_rhs = []
        coeffs = []

        for j in range(len(names) - 1):
            if A_aug_rref[i, j] != 0 and j != pivots[i]:
                coeffs.append(-A_aug_rref[i, j])
                names_rhs.append(names[j])

        coeffs.append(A_aug_rref[i, -1])

        _equalities.append((name_lhs, names_rhs, coeffs))

    trafo = AffineTransform(_equalities)
    # remove remaining dependencies of eliminated inputs from the problem
    _domain = remove_eliminated_inputs(_domain, trafo)
    return _domain, trafo


def check_domain_for_reduction(domain: Domain) -> bool:
    """Check if the reduction can be applied or if a trivial case is present.

    Args:
        domain (Domain): Domain to be checked.
    Returns:
        bool: True if reducable, else False.
    """
    # are there any constraints?
    if len(domain.constraints) == 0:
        return False

    # are there any linear equality constraints?
    linear_equalities = domain.constraints.get(LinearEqualityConstraint)
    if len(linear_equalities) == 0:
        return False

    # are there no NChooseKConstraint constraints?
    if len(domain.constraints.get([NChooseKConstraint])) > 0:
        return False

    # are there continuous inputs
    continuous_inputs = domain.get_features(ContinuousInput)
    if len(continuous_inputs) == 0:
        return False

    # check that equality constraints only contain continuous inputs
    for c in linear_equalities:
        assert isinstance(c, LinearConstraint)
        for feat in c.features:
            if feat not in domain.get_feature_keys(ContinuousInput):
                return False
    return True


def check_existence_of_solution(A_aug):
    """Given an augmented coefficient matrix this function determines the existence (and uniqueness) of solution using the rank theorem."""
    A = A_aug[:, :-1]
    b = A_aug[:, -1]
    len_inputs = np.shape(A)[1]

    # catch special cases
    rk_A_aug = np.linalg.matrix_rank(A_aug)
    rk_A = np.linalg.matrix_rank(A)

    if rk_A == rk_A_aug:
        if rk_A < len_inputs:
            return  # all good
        else:
            x = np.linalg.solve(A, b)
            raise Exception(
                f"There is a unique solution x for the linear equality constraints: x={x}"
            )
    elif rk_A < rk_A_aug:
        raise Exception(
            "There is no solution fulfilling the linear equality constraints."
        )


def remove_eliminated_inputs(domain: Domain, transform: AffineTransform) -> Domain:
    """Eliminates remaining occurences of eliminated inputs in linear constraints.

    Args:
        domain (Domain): Domain in which the linear constraints should be purged.
        transform (AffineTransform): Affine transformation object that defines the obsolete features.

    Raises:
        ValueError: If feature occurs in a constraint different from a linear one.

    Returns:
        Domain: Purged domain.
    """
    inputs_names = domain.get_feature_keys()
    M = len(inputs_names)

    # write the equalities for the backtransformation into one matrix
    inputs_dict = {inputs_names[i]: i for i in range(M)}

    # build up dict from domain.equalities e.g. {"xi1": [coeff(xj1), ..., coeff(xjn)], ... "xik":...}
    coeffs_dict = {}
    for e in transform.equalities:
        coeffs = np.zeros(M + 1)
        for j, name in enumerate(e[1]):
            coeffs[inputs_dict[name]] = e[2][j]
        coeffs[-1] = e[2][-1]
        coeffs_dict[e[0]] = coeffs

    constraints = []
    for c in domain.constraints.get():
        # Nonlinear constraints not supported
        if not isinstance(c, LinearConstraint):
            raise ValueError(
                "Elimination of variables is only supported for LinearEquality and LinearInequality constraints."
            )

        # no changes, if the constraint does not contain eliminated inputs
        elif all(name in inputs_names for name in c.features):
            constraints.append(c)

        # remove inputs from the constraint that were eliminated from the inputs before
        else:
            totally_removed = False
            _features = np.array(inputs_names)
            _rhs = c.rhs

            # create new lhs and rhs from the old one and knowledge from problem._equalities
            _coefficients = np.zeros(M)
            for j, name in enumerate(c.features):
                if name in inputs_names:
                    _coefficients[inputs_dict[name]] += c.coefficients[j]
                else:
                    _coefficients += c.coefficients[j] * coeffs_dict[name][:-1]
                    _rhs -= c.coefficients[j] * coeffs_dict[name][-1]

            _features = _features[np.abs(_coefficients) > 1e-16]
            _coefficients = _coefficients[np.abs(_coefficients) > 1e-16]
            _c = None
            if isinstance(c, LinearEqualityConstraint):
                if len(_features) > 1:
                    _c = LinearEqualityConstraint(
                        features=_features.tolist(),
                        coefficients=_coefficients.tolist(),
                        rhs=_rhs,
                    )
                elif len(_features) == 0:
                    totally_removed = True
                else:
                    feat: ContinuousInput = ContinuousInput(
                        **domain.get_feature(_features[0]).dict()
                    )
                    feat.bounds = (_coefficients[0], _coefficients[0])
                    totally_removed = True
            else:
                if len(_features) > 1:
                    _c = LinearInequalityConstraint(
                        features=_features.tolist(),
                        coefficients=_coefficients.tolist(),
                        rhs=_rhs,
                    )
                elif len(_features) == 0:
                    totally_removed = True
                else:
                    feat = cast(ContinuousInput, domain.get_feature(_features[0]))
                    adjust_boundary(feat, _coefficients[0], _rhs)
                    totally_removed = True

            # check if constraint is always fulfilled/not fulfilled
            if not totally_removed:
                assert _c is not None
                if len(_c.features) == 0 and _c.rhs >= 0:
                    pass
                elif len(_c.features) == 0 and _c.rhs < 0:
                    raise Exception("Linear constraints cannot be fulfilled.")
                elif np.isinf(_c.rhs):
                    pass
                else:
                    constraints.append(_c)
    domain.constraints = Constraints(constraints=constraints)
    return domain


def rref(A: np.ndarray, tol: float = 1e-8) -> Tuple[np.ndarray, List[int]]:
    """Computes the reduced row echelon form of a Matrix

    Args:
        A (ndarray): 2d array representing a matrix.
        tol (float, optional): tolerance for rounding to 0. Defaults to 1e-8.

    Returns:
        (A_rref, pivots), where A_rref is the reduced row echelon form of A and pivots
        is a numpy array containing the pivot columns of A_rref
    """
    A = np.array(A, dtype=np.float64)
    n, m = np.shape(A)

    col = 0
    row = 0
    pivots = []

    for col in range(m):
        # does a pivot element exist?
        if all(np.abs(A[row:, col]) < tol):
            pass
        # if yes: start elimination
        else:
            pivots.append(col)
            max_row = np.argmax(np.abs(A[row:, col])) + row
            # switch to most stable row
            A[[row, max_row], :] = A[[max_row, row], :]  # type: ignore
            # normalize row
            A[row, :] /= A[row, col]
            # eliminate other elements from column
            for r in range(n):
                if r != row:
                    A[r, :] -= A[r, col] / A[row, col] * A[row, :]
            row += 1

    prec = int(-np.log10(tol))
    return np.round(A, prec), pivots


def adjust_boundary(feature: ContinuousInput, coef: float, rhs: float):
    """Adjusts the boundaries of a feature.

    Args:
        feature (ContinuousInput): Feature to be adjusted.
        coef (float): Coefficient.
        rhs (float): Right-hand-side of the constraint.
    """
    boundary = rhs / coef
    if coef > 0:
        if boundary > feature.lower_bound:
            feature.bounds = (boundary, feature.upper_bound)
    else:
        if boundary < feature.upper_bound:
            feature.bounds = (feature.lower_bound, boundary)



import sys

import numpy as np
import pytest
from scipy.optimize import LinearConstraint, NonlinearConstraint

from bofire.data_models.constraints.api import (
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NChooseKConstraint,
    NonlinearEqualityConstraint,
    NonlinearInequalityConstraint,
)
from bofire.data_models.domain.api import Domain
from bofire.data_models.features.api import (
    ContinuousInput,
    ContinuousOutput,
    DiscreteInput,
)
from bofire.strategies.doe.utils import (
    ConstraintWrapper,
    a_optimality,
    check_nchoosek_constraints_as_bounds,
    constraints_as_scipy_constraints,
    d_optimality,
    g_optimality,
    get_formula_from_string,
    metrics,
    n_zero_eigvals,
    nchoosek_constraints_as_bounds,
)


def get_formula_from_string_recursion_limit():
    # save recursion limit
    recursion_limit = sys.getrecursionlimit()

    # get formula for very large model
    model = ""
    for i in range(350):
        model += f"x{i} + "
    model = model[:-3]
    model = get_formula_from_string(model_type=model)

    terms = [f"x{i}" for i in range(350)]
    terms.append("1")

    for i in range(351):
        assert np.array(model, dtype=str)[i] in terms
        assert terms[i] in model

    assert recursion_limit == sys.getrecursionlimit()


def test_get_formula_from_string():
    domain = Domain.from_lists(
        inputs=[ContinuousInput(key=f"x{i}", bounds=(0, 1)) for i in range(3)],
        outputs=[ContinuousOutput(key="y")],
    )

    # linear model
    terms = ["1", "x0", "x1", "x2"]
    model_formula = get_formula_from_string(domain=domain, model_type="linear")
    assert all(term in terms for term in model_formula)
    assert all(term in np.array(model_formula, dtype=str) for term in terms)

    # linear and interaction
    terms = ["1", "x0", "x1", "x2", "x0:x1", "x0:x2", "x1:x2"]
    model_formula = get_formula_from_string(
        domain=domain, model_type="linear-and-interactions"
    )
    assert all(term in terms for term in model_formula)
    assert all(term in np.array(model_formula, dtype=str) for term in terms)

    # linear and quadratic
    terms = ["1", "x0", "x1", "x2", "x0**2", "x1**2", "x2**2"]
    model_formula = get_formula_from_string(
        domain=domain, model_type="linear-and-quadratic"
    )
    assert all(term in terms for term in model_formula)
    assert all(term in np.array(model_formula, dtype=str) for term in terms)

    # fully quadratic
    terms = [
        "1",
        "x0",
        "x1",
        "x2",
        "x0:x1",
        "x0:x2",
        "x1:x2",
        "x0**2",
        "x1**2",
        "x2**2",
    ]
    model_formula = get_formula_from_string(domain=domain, model_type="fully-quadratic")
    assert all(term in terms for term in model_formula)
    assert all(term in np.array(model_formula, dtype=str) for term in terms)

    # custom model
    terms_lhs = ["y"]
    terms_rhs = ["1", "x0", "x0**2", "x0:x1"]
    model_formula = get_formula_from_string(
        domain=domain,
        model_type="y ~ 1 + x0 + x0:x1 + {x0**2}",
        rhs_only=False,
    )
    assert all(term in terms_lhs for term in model_formula.lhs)
    assert all(term in str(model_formula.lhs) for term in terms_lhs)
    assert all(term in terms_rhs for term in model_formula.rhs)
    assert all(term in np.array(model_formula.rhs, dtype=str) for term in terms_rhs)

    # get formula without model: valid input
    model = "x1 + x2 + x3"
    model = get_formula_from_string(model_type=model)
    assert str(model) == "1 + x1 + x2 + x3"

    # get formula without model: invalid input
    with pytest.raises(AssertionError):
        model = get_formula_from_string("linear")

    # get formula for very large model
    model = ""
    for i in range(350):
        model += f"x{i} + "
    model = model[:-3]
    model = get_formula_from_string(model_type=model)

    terms = [f"x{i}" for i in range(350)]
    terms.append("1")

    for i in range(351):
        assert list(model)[i] in terms
        assert terms[i] in np.array(model, dtype=str)


def test_n_zero_eigvals_unconstrained():
    # 5 continous
    domain = Domain.from_lists(
        inputs=[ContinuousInput(key=f"x{i+1}", bounds=(0, 100)) for i in range(5)],
        outputs=[ContinuousOutput(key="y")],
    )

    assert n_zero_eigvals(domain=domain, model_type="linear") == 0
    assert n_zero_eigvals(domain=domain, model_type="linear-and-quadratic") == 0
    assert n_zero_eigvals(domain=domain, model_type="linear-and-interactions") == 0
    assert n_zero_eigvals(domain=domain, model_type="fully-quadratic") == 0


def test_n_zero_eigvals_constrained():
    # 3 continuous & 2 discrete inputs, 1 mixture constraint
    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(key="x1", bounds=(0, 100)),
            ContinuousInput(key="x2", bounds=(0, 100)),
            ContinuousInput(key="x3", bounds=(0, 100)),
            DiscreteInput(key="discrete1", values=[0, 1, 5]),
            DiscreteInput(key="discrete2", values=[0, 1]),
        ],
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            LinearEqualityConstraint(
                features=["x1", "x2", "x3"], coefficients=[1, 1, 1], rhs=1
            )
        ],
    )

    # Comment DL: in the test in the doe package discrete was sampled as continuous
    # thus there was one degree of freedom more if quadratic terms where added.
    # Here, discretes are sampled within their respective domain, thus discrete2==discrete2**2 always
    # thus we have one degree of freedom less.
    assert n_zero_eigvals(domain, "linear") == 1
    assert n_zero_eigvals(domain, "linear-and-quadratic") == 2
    assert n_zero_eigvals(domain, "linear-and-interactions") == 3
    assert n_zero_eigvals(domain, "fully-quadratic") == 7

    # TODO: NChooseK?


def test_number_of_model_terms():
    # 5 continous inputs
    domain = Domain.from_lists(
        inputs=[ContinuousInput(key=f"x{i}", bounds=(0, 1)) for i in range(5)],
        outputs=[ContinuousOutput(key="y")],
    )

    formula = get_formula_from_string(domain=domain, model_type="linear")
    assert len(formula) == 6

    formula = get_formula_from_string(domain=domain, model_type="linear-and-quadratic")
    assert len(formula) == 11

    formula = get_formula_from_string(
        domain=domain, model_type="linear-and-interactions"
    )
    assert len(formula) == 16

    formula = get_formula_from_string(domain=domain, model_type="fully-quadratic")
    assert len(formula) == 21

    # 3 continuous & 2 discrete inputs
    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(key="x1", bounds=(0, 100)),
            ContinuousInput(key="x2", bounds=(0, 100)),
            ContinuousInput(key="x3", bounds=(0, 100)),
            DiscreteInput(key="discrete1", values=[0, 1, 5]),
            DiscreteInput(key="discrete2", values=[0, 1]),
        ],
        outputs=[ContinuousOutput(key="y")],
    )

    formula = get_formula_from_string(domain=domain, model_type="linear")
    assert len(formula) == 6

    formula = get_formula_from_string(domain=domain, model_type="linear-and-quadratic")
    assert len(formula) == 11

    formula = get_formula_from_string(
        domain=domain, model_type="linear-and-interactions"
    )
    assert len(formula) == 16

    formula = get_formula_from_string(domain=domain, model_type="fully-quadratic")
    assert len(formula) == 21


def test_constraints_as_scipy_constraints():
    # test domains from the paper "The construction of D- and I-optimal designs for
    # mixture experiments with linear constraints on the components" by R. Coetzer and
    # L. M. Haines.
    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(key=f"x{1}", bounds=(0, 1)),
            ContinuousInput(key=f"x{2}", bounds=(0.1, 1)),
            ContinuousInput(key=f"x{3}", bounds=(0, 0.6)),
        ],
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            LinearEqualityConstraint(
                features=["x1", "x2", "x3"], coefficients=[1, 1, 1], rhs=1
            ),
            LinearInequalityConstraint(
                features=["x1", "x2"], coefficients=[5, 4], rhs=3.9
            ),
            LinearInequalityConstraint(
                features=["x1", "x2"], coefficients=[-20, 5], rhs=-3
            ),
        ],
    )

    n_experiments = 2

    constraints = constraints_as_scipy_constraints(domain, n_experiments)

    for c in constraints:
        assert isinstance(c, LinearConstraint)
        assert c.A.shape == (n_experiments, len(domain.inputs) * n_experiments)
        assert len(c.lb) == n_experiments
        assert len(c.ub) == n_experiments

    A = np.array([[1, 1, 1, 0, 0, 0], [0, 0, 0, 1, 1, 1]]) / np.sqrt(3)
    lb = np.array([1, 1]) / np.sqrt(3)
    ub = np.array([1, 1]) / np.sqrt(3)
    assert np.allclose(constraints[0].A, A)
    assert np.allclose(constraints[0].lb, lb)
    assert np.allclose(constraints[0].ub, ub)

    lb = -np.inf * np.ones(n_experiments)
    ub = 3.9 / np.linalg.norm([5, 4]) * np.ones(n_experiments)
    assert np.allclose(constraints[1].lb, lb)
    assert np.allclose(constraints[1].ub, ub)

    # domain with nonlinear constraints
    domain = Domain.from_lists(
        inputs=[ContinuousInput(key=f"x{i+1}", bounds=(0, 1)) for i in range(3)],
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            NonlinearEqualityConstraint(
                expression="x1**2 + x2**2 - 1", features=["x1", "x2", "x3"]
            ),
            NonlinearInequalityConstraint(
                expression="x1**2 + x2**2 - 1", features=["x1", "x2", "x3"]
            ),
        ],
    )

    constraints = constraints_as_scipy_constraints(domain, n_experiments)

    for c in constraints:
        assert isinstance(c, NonlinearConstraint)
        assert len(c.lb) == n_experiments
        assert len(c.ub) == n_experiments
        assert np.allclose(c.fun(np.array([1, 1, 1, 1, 1, 1])), [1, 1])

    # TODO NChooseKConstraint requires input lower_bounds to be 0.
    # can we lift this requirment?

    inputs = [ContinuousInput(key=f"x{i}", bounds=(0, 1)) for i in range(4)]

    domain = Domain.from_lists(
        inputs=inputs,
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            NChooseKConstraint(
                features=[i.key for i in inputs],
                max_count=2,
                min_count=0,
                none_also_valid=True,
            )
        ],
    )
    n_experiments = 1

    constraints = constraints_as_scipy_constraints(
        domain, n_experiments, ignore_nchoosek=True
    )
    assert len(constraints) == 0

    constraints = constraints_as_scipy_constraints(
        domain, n_experiments, ignore_nchoosek=False
    )
    assert len(constraints) == 1
    assert isinstance(constraints[0], NonlinearConstraint)
    assert np.allclose(
        constraints[0].fun(np.array([1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0])), [2, 0, 0]
    )


def test_ConstraintWrapper():
    # define domain with all types of constraints
    domain = Domain.from_lists(
        inputs=[ContinuousInput(key=f"x{i+1}", bounds=(0, 1)) for i in range(4)],
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            LinearEqualityConstraint(
                features=["x1", "x2", "x3", "x4"], coefficients=[1, 1, 1, 1], rhs=1
            ),
            LinearInequalityConstraint(
                features=["x1", "x2", "x3", "x4"], coefficients=[1, 1, 1, 1], rhs=1
            ),
            NonlinearEqualityConstraint(
                expression="x1**2 + x2**2 + x3**2 + x4**2 - 1",
                features=["x1", "x2", "x3", "x4"],
                jacobian_expression="[2*x1, 2*x2, 2*x3, 2*x4]",
            ),
            NonlinearInequalityConstraint(
                expression="x1**2 + x2**2 + x3**2 + x4**2 - 1",
                features=["x1", "x2", "x3", "x4"],
                jacobian_expression="[2*x1, 2*x2, 2*x3, 2*x4]",
            ),
            NonlinearEqualityConstraint(
                expression="x1**2 + x4**2 - 1",
                features=["x1", "x4"],
                jacobian_expression="[2*x1, 2*x4]",
            ),
        ],
    )

    x = np.array([[1, 1, 1, 1], [0.5, 0.5, 0.5, 0.5], [3, 2, 1, 0]]).flatten()

    # linear equality
    c = ConstraintWrapper(domain.constraints[0], domain, n_experiments=3)
    assert np.allclose(c(x), np.array([1.5, 0.5, 2.5]))
    assert np.allclose(
        c.jacobian(x),
        0.5
        * np.array(
            [
                [1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1],
            ]
        ),
    )

    # linear inequality
    c = ConstraintWrapper(domain.constraints[1], domain, n_experiments=3)
    assert np.allclose(c(x), np.array([1.5, 0.5, 2.5]))
    assert np.allclose(
        c.jacobian(x),
        0.5
        * np.array(
            [
                [1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1],
            ]
        ),
    )

    # nonlinear equality
    c = ConstraintWrapper(domain.constraints[2], domain, n_experiments=3)
    assert np.allclose(c(x), np.array([3, 0, 13]))
    assert np.allclose(
        c.jacobian(x),
        np.array(
            [
                [2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 6, 4, 2, 0],
            ]
        ),
    )

    # nonlinear inequality
    c = ConstraintWrapper(domain.constraints[3], domain, n_experiments=3)
    assert np.allclose(c(x), np.array([3, 0, 13]))
    assert np.allclose(
        c.jacobian(x),
        np.array(
            [
                [2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 6, 4, 2, 0],
            ]
        ),
    )

    # constraint not containing all inputs from domain
    c = ConstraintWrapper(domain.constraints[4], domain, n_experiments=3)
    assert np.allclose(c(x), np.array([1, -0.5, 8]))
    assert np.allclose(
        c.jacobian(x),
        np.array(
            [
                [2, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 0],
            ]
        ),
    )


def test_d_optimality():
    # define model matrix: full rank
    X = np.array(
        [
            [1, 1, 0, 0],
            [1, 0, 1, 0],
            [1, 0, 0, 1],
            [1, 0, 0, 0],
        ]
    )
    assert np.allclose(d_optimality(X), np.linalg.slogdet(X.T @ X)[1])

    # define model matrix: not full rank
    X = np.array(
        [
            [1, 1, 0, 0],
            [1, 0, 1, 0],
            [1, 0, 0, 1],
            [1, 1 / 3, 1 / 3, 1 / 3],
        ]
    )
    assert np.allclose(d_optimality(X), np.sum(np.log(np.linalg.eigvalsh(X.T @ X)[1:])))


def test_a_optimality():
    # define model matrix: full rank
    X = np.array(
        [
            [1, 1, 0, 0],
            [1, 0, 1, 0],
            [1, 0, 0, 1],
            [1, 0, 0, 0],
        ]
    )
    assert np.allclose(a_optimality(X), np.sum(1 / (np.linalg.eigvalsh(X.T @ X))))

    # define model matrix: not full rank
    X = np.array(
        [
            [1, 1, 0, 0],
            [1, 0, 1, 0],
            [1, 0, 0, 1],
            [1, 1 / 3, 1 / 3, 1 / 3],
        ]
    )
    assert np.allclose(a_optimality(X), np.sum(1 / (np.linalg.eigvalsh(X.T @ X)[1:])))


def test_g_optimality():
    # define model matrix and domain: no constraints
    X = np.array(
        [
            [1, 0, 0, 0],
            [0, 0.1, 0, 0],
            [0, 0, 0.1, 0],
            [0, 0, 0, 0.1],
        ]
    )
    assert np.allclose(g_optimality(X), 1)


def test_metrics():
    # define model matrix
    X = np.array(
        [
            [1, 1, 0, 0],
            [1, 0, 1, 0],
            [1, 0, 0, 1],
            [1, 0, 0, 0],
        ]
    )

    m = metrics(X)
    assert np.allclose(m["A-optimality"], a_optimality(X))
    assert np.allclose(m["D-optimality"], d_optimality(X))
    assert np.allclose(m["G-optimality"], g_optimality(X))


def test_check_nchoosek_constraints_as_bounds():
    # define domain: possible to formulate as bounds, no NChooseK constraints
    domain = Domain.from_lists(
        inputs=[ContinuousInput(key=f"x{i+1}", bounds=(0, 1)) for i in range(4)],
        outputs=[ContinuousOutput(key="y")],
    )
    check_nchoosek_constraints_as_bounds(domain)

    domain = Domain.from_lists(
        inputs=[ContinuousInput(key=f"x{i+1}", bounds=(-1, 1)) for i in range(4)],
        outputs=[ContinuousOutput(key="y")],
        constraints=[],
    )
    check_nchoosek_constraints_as_bounds(domain)

    domain = Domain.from_lists(
        inputs=[ContinuousInput(key=f"x{i+1}", bounds=(-np.inf, 1)) for i in range(4)],
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            LinearEqualityConstraint(features=["x1", "x2"], coefficients=[1, 1], rhs=0)
        ],
    )
    check_nchoosek_constraints_as_bounds(domain)

    # n-choose-k constraints when variables can be negative
    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(key=f"x{1}", bounds=(0, 1)),
            ContinuousInput(key=f"x{2}", bounds=(-1, 1)),
            ContinuousInput(key=f"x{3}", bounds=(-2, 1)),
            ContinuousInput(key=f"x{4}", bounds=(-3, 1)),
        ],
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            LinearEqualityConstraint(features=["x1", "x2"], coefficients=[1, 1], rhs=0),
            LinearInequalityConstraint(
                features=["x3", "x4"], coefficients=[1, 1], rhs=0
            ),
            NChooseKConstraint(
                features=["x1", "x2"], max_count=1, min_count=0, none_also_valid=True
            ),
            NChooseKConstraint(
                features=["x3", "x4"], max_count=1, min_count=0, none_also_valid=True
            ),
        ],
    )
    check_nchoosek_constraints_as_bounds(domain)

    # It should be allowed to have n-choose-k constraints when 0 is not in the bounds.
    domain = Domain.from_lists(
        inputs=[ContinuousInput(key=f"x{i+1}", bounds=(0.1, 1)) for i in range(4)],
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            NChooseKConstraint(
                features=["x1", "x2"],
                max_count=1,
                min_count=0,
                none_also_valid=True,
            ),
        ],
    )
    with pytest.raises(ValueError):
        check_nchoosek_constraints_as_bounds(domain)  # FIXME: should be allowed

    # It should be allowed to have n-choose-k constraints when 0 is not in the bounds.
    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(key=f"x{1}", bounds=(-1, -0.1)),
            ContinuousInput(key=f"x{2}", bounds=(-1, -0.1)),
            ContinuousInput(key=f"x{3}", bounds=(-1, -0.1)),
            ContinuousInput(key=f"x{4}", bounds=(-1, -0.1)),
        ],
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            NChooseKConstraint(
                features=["x1", "x2"],
                max_count=1,
                min_count=0,
                none_also_valid=True,
            ),
        ],
    )
    with pytest.raises(ValueError):
        check_nchoosek_constraints_as_bounds(domain)  # FIXME: should be allowed

    # Not allowed: names parameters of two NChooseK overlap
    domain = Domain.from_lists(
        inputs=[ContinuousInput(key=f"x{i+1}", bounds=(0, 1)) for i in range(4)],
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            NChooseKConstraint(
                features=["x1", "x2"],
                max_count=1,
                min_count=0,
                none_also_valid=True,
            ),
            NChooseKConstraint(
                features=["x2", "x3", "x4"],
                max_count=2,
                min_count=0,
                none_also_valid=True,
            ),
        ],
    )
    with pytest.raises(ValueError):
        check_nchoosek_constraints_as_bounds(domain)


def test_nchoosek_constraints_as_bounds():
    # define domain: no NChooseK constraints
    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(
                key=f"x{i+1}",
                bounds=(-1, 1),
            )
            for i in range(5)
        ],
        outputs=[ContinuousOutput(key="y")],
    )
    bounds = nchoosek_constraints_as_bounds(domain, n_experiments=4)
    assert len(bounds) == 20
    _bounds = [
        (p.lower_bound, p.upper_bound)
        for p in domain.inputs
        if isinstance(p, ContinuousInput)
    ] * 4
    for i in range(20):
        assert _bounds[i] == bounds[i]

    # define domain: with NChooseK constraints
    # define domain: no NChooseK constraints
    # domain = Domain(
    #     inputs=[
    #         ContinuousInput(key=f"x{i+1}", bounds=(-1, 1),)
    #         for i in range(5)
    #     ],
    #     outputs=[ContinuousOutput(key="y")],
    #     constraints=[opti.NChooseK(["x1", "x2", "x3"], max_active=1)],
    # )
    # np.random.seed(1)
    # bounds = nchoosek_constraints_as_bounds(domain, n_experiments=4)
    # _bounds = [
    #     (0.0, 0.0),
    #     (0.0, 0.0),
    #     (-1.0, 1.0),
    #     (-1.0, 1.0),
    #     (-1.0, 1.0),
    #     (-1.0, 1.0),
    #     (0.0, 0.0),
    #     (0.0, 0.0),
    #     (-1.0, 1.0),
    #     (-1.0, 1.0),
    #     (0.0, 0.0),
    #     (-1.0, 1.0),
    #     (0.0, 0.0),
    #     (-1.0, 1.0),
    #     (-1.0, 1.0),
    #     (0.0, 0.0),
    #     (0.0, 0.0),
    #     (-1.0, 1.0),
    #     (-1.0, 1.0),
    #     (-1.0, 1.0),
    # ]
    # assert len(bounds) == 20
    # for i in range(20):
    #     assert _bounds[i] == bounds[i]



from typing import Any

import pytest

from bofire.data_models.base import filter_by_attribute, filter_by_class


class A:
    pass


class A1(A):
    pass


class A11(A1):
    pass


class A12(A1):
    pass


class A2(A):
    pass


class A21(A2):
    pass


class A211(A21):
    pass


class B:
    pass


class B1(B):
    pass


class B11(B1):
    pass


class B12(B1):
    pass


class B2(B):
    pass


class B21(B2):
    pass


class B211(B21):
    pass


class C:
    def __init__(self, attribute: Any) -> None:
        self.attribute = attribute


a = A()
a1 = A1()
a11 = A11()
a12 = A12()
a2 = A2()
a21 = A21()
a211 = A211()

b = B()
b1 = B1()
b11 = B11()
b12 = B12()
b2 = B2()
b21 = B21()
b211 = B211()


c = C(attribute=b)
c1 = C(attribute=b1)
c11 = C(attribute=b11)
c12 = C(attribute=b12)
c2 = C(attribute=b2)
c21 = C(attribute=b21)
c211 = C(attribute=b211)
c3 = C(attribute=None)

data = [
    a,
    a1,
    a11,
    a12,
    a2,
    a21,
    a211,
]

data2 = [
    a,
    b,
    c,
    c1,
    c11,
    c12,
    c2,
    c21,
    c211,
    c3,
]


@pytest.mark.parametrize(
    "data,includes,expected",
    [
        (data, [A], data),
        (data, [A, A1, A21], data),
        (data, [A1], [a1, a11, a12]),
        (data, [A11], [a11]),
        (data, [A2], [a2, a21, a211]),
    ],
)
def test_filter_by_class(data, includes, expected):
    res = filter_by_class(data, includes)
    print("got:", [type(x).__name__ for x in res])
    print("expected:", [type(x).__name__ for x in expected])
    assert set(res) == set(expected)


@pytest.mark.parametrize(
    "data,includes,expected",
    [
        (data2, [B], [c, c1, c11, c12, c2, c21, c211]),
        (data, [B], []),
        (data, [B], []),
        (data2, [B, B1, B2], [c, c1, c11, c12, c2, c21, c211]),
        (data2, [B1], [c1, c11, c12]),
        (data2, [B11], [c11]),
        (data2, [B2], [c2, c21, c211]),
    ],
)
def test_filter_by_attribute(data, includes, expected):
    res = filter_by_attribute(
        data, attribute_getter=lambda of: of.attribute, includes=includes
    )
    print("got:", [type(x.attribute).__name__ for x in res])
    print("expected:", [type(x.attribute).__name__ for x in expected])
    assert set(res) == set(expected)


@pytest.mark.parametrize(
    "data,excludes,expected",
    [
        (data, [A], []),
        (data, [A2], [a, a1, a11, a12]),
    ],
)
def test_filter_by_class_only_exclude(data, excludes, expected):
    res = filter_by_class(data, excludes=excludes)
    print("got:", [type(x).__name__ for x in res])
    print("expected:", [type(x).__name__ for x in expected])
    assert set(res) == set(expected)


@pytest.mark.parametrize(
    "data,excludes,expected",
    [
        (data2, [B2], [c, c1, c11, c12, c3]),
        (data2, [B], [c3]),
    ],
)
def test_filter_by_attribute_only_exclude(data, excludes, expected):
    res = filter_by_attribute(
        data, attribute_getter=lambda of: of.attribute, excludes=excludes
    )
    print("got:", [type(x.attribute).__name__ for x in res])
    print("expected:", [type(x.attribute).__name__ for x in expected])
    assert set(res) == set(expected)


@pytest.mark.parametrize(
    "data,includes,expected",
    [
        (data, [A], [a]),
        (data, [A, A1, A21], [a, a1, a21]),
        (data, [A1], [a1]),
        (data, [A11], [a11]),
        (data, [A2], [a2]),
    ],
)
def test_filter_by_class_exact(data, includes, expected):
    res = filter_by_class(data, includes, exact=True)
    print("got:", [type(x).__name__ for x in res])
    print("expected:", [type(x).__name__ for x in expected])
    assert set(res) == set(expected)


@pytest.mark.parametrize(
    "data,includes,expected",
    [
        (data2, [B], [c]),
        (data2, [B, B1, B21], [c, c1, c21]),
        (data2, [B1], [c1]),
        (data2, [B11], [c11]),
        (data2, [B2], [c2]),
    ],
)
def test_filter_by_attribute_exact(data, includes, expected):
    res = filter_by_attribute(
        data, attribute_getter=lambda of: of.attribute, includes=includes, exact=True
    )
    print("got:", [type(x.attribute).__name__ for x in res])
    print("expected:", [type(x.attribute).__name__ for x in expected])
    assert set(res) == set(expected)


@pytest.mark.parametrize(
    "data,includes,exclude,expected",
    [
        (data, [A], [], data),
        (data, [A1], [A11], [a1, a12]),
        (data, [A], [A21], [a, a1, a11, a12, a2]),
        (data, [A2], [A1], [a2, a21, a211]),
    ],
)
def test_filter_by_class_exclude(data, includes, exclude, expected):
    res = filter_by_class(data, includes, exclude)
    print("got:", [type(x).__name__ for x in res])
    print("expected:", [type(x).__name__ for x in expected])
    assert set(res) == set(expected)


@pytest.mark.parametrize(
    "data,includes,exclude,expected",
    [
        (data2, [B], [], [c, c1, c11, c12, c2, c21, c211]),
        (data2, [B1], [B11], [c1, c12]),
        (data2, [B], [B21], [c, c1, c11, c12, c2]),
        (data2, [B2], [B1], [c2, c21, c211]),
    ],
)
def test_filter_by_class_exclude_attribute(data, includes, exclude, expected):
    res = filter_by_attribute(
        data,
        attribute_getter=lambda of: of.attribute,
        includes=includes,
        excludes=exclude,
    )
    print("got:", [type(x.attribute).__name__ for x in res])
    print("expected:", [type(x.attribute).__name__ for x in expected])
    assert set(res) == set(expected)


@pytest.mark.parametrize(
    "data,includes,exclude,expected",
    [
        (data, [A], [], [a]),
        (data, [A1], [A11], [a1]),
        (data, [A1], [], [a1]),
        (data, [A1, A11], [], [a1, a11]),
        (data, [A, A2, A21], [A211], [a, a2, a21]),
    ],
)
def test_filter_by_class_exclude_exact(data, includes, exclude, expected):
    res = filter_by_class(data, includes, exclude, exact=True)
    print("got:", [type(x).__name__ for x in res])
    print("expected:", [type(x).__name__ for x in expected])
    assert set(res) == set(expected)


@pytest.mark.parametrize(
    "data,includes,exclude,expected",
    [
        (data2, [B], [], [c]),
        (data2, [B1], [B11], [c1]),
        (data2, [B1], [], [c1]),
        (data2, [B1, B11], [], [c1, c11]),
        (data2, [B, B2, B21], [B211], [c, c2, c21]),
    ],
)
def test_filter_by_attribute_exclude_exact(data, includes, exclude, expected):
    res = filter_by_attribute(
        data,
        attribute_getter=lambda of: of.attribute,
        includes=includes,
        excludes=exclude,
        exact=True,
    )
    print("got:", [type(x.attribute).__name__ for x in res])
    print("expected:", [type(x.attribute).__name__ for x in expected])
    assert set(res) == set(expected)


@pytest.mark.parametrize(
    "includes, excludes",
    [
        ([], []),
        ([], None),
        (None, None),
        (None, []),
    ],
)
def test_filter_by_class_no_cls(includes, excludes):
    with pytest.raises(ValueError):
        filter_by_class(data, includes=includes, excludes=excludes)


@pytest.mark.parametrize(
    "includes, excludes",
    [
        ([A], [A]),
        ([A2], [A2]),
        ([A12], [A, A12]),
        ([A11, A21], [A12, A21]),
    ],
)
def test_filter_by_class_no_intersection(includes, excludes):
    with pytest.raises(ValueError):
        filter_by_class(data, includes, excludes)



from typing import List, Literal, Type, Union

from pydantic import Field, validator
from typing_extensions import Annotated

from bofire.data_models.base import BaseModel
from bofire.data_models.constraints.api import Constraint
from bofire.data_models.features.api import Feature
from bofire.data_models.strategies.doe import DoEStrategy
from bofire.data_models.strategies.predictives.qehvi import QehviStrategy
from bofire.data_models.strategies.predictives.qnehvi import QnehviStrategy
from bofire.data_models.strategies.predictives.qparego import QparegoStrategy
from bofire.data_models.strategies.predictives.sobo import (
    AdditiveSoboStrategy,
    MultiplicativeSoboStrategy,
    SoboStrategy,
)
from bofire.data_models.strategies.random import RandomStrategy
from bofire.data_models.strategies.samplers.polytope import PolytopeSampler
from bofire.data_models.strategies.samplers.rejection import RejectionSampler
from bofire.data_models.strategies.stepwise.conditions import (
    AlwaysTrueCondition,
    CombiCondition,
    NumberOfExperimentsCondition,
)
from bofire.data_models.strategies.strategy import Strategy

AnyStrategy = Union[
    SoboStrategy,
    AdditiveSoboStrategy,
    MultiplicativeSoboStrategy,
    QehviStrategy,
    QnehviStrategy,
    QparegoStrategy,
    PolytopeSampler,
    RejectionSampler,
    RandomStrategy,
    DoEStrategy,
]

AnyCondition = Union[NumberOfExperimentsCondition, CombiCondition, AlwaysTrueCondition]


class Step(BaseModel):
    type: Literal["Step"] = "Step"
    strategy_data: AnyStrategy
    condition: AnyCondition
    max_parallelism: Annotated[int, Field(ge=-1)]  # -1 means no restriction at all


class StepwiseStrategy(Strategy):
    type: Literal["StepwiseStrategy"] = "StepwiseStrategy"
    steps: Annotated[List[Step], Field(min_items=2)]

    @validator("steps")
    def validate_steps(cls, v: List[Step], values):
        for i, step in enumerate(v):
            if step.strategy_data.domain != values["domain"]:
                raise ValueError(
                    f"Domain of step {i} is incompatible to domain of StepwiseStrategy."
                )
            if i < len(v) - 1 and isinstance(step.condition, AlwaysTrueCondition):
                raise ValueError(
                    "`AlwaysTrueCondition` is only allowed for the last step."
                )
        return v

    @classmethod
    def is_feature_implemented(cls, my_type: Type[Feature]) -> bool:
        return True

    @classmethod
    def is_constraint_implemented(cls, my_type: Type[Constraint]) -> bool:
        return True



from typing import Dict, Literal, Optional, Type

from pydantic import validator

from bofire.data_models.features.api import CategoricalOutput, Feature
from bofire.data_models.objectives.api import (
    MaximizeObjective,
    MinimizeObjective,
    Objective,
)
from bofire.data_models.strategies.predictives.multiobjective import (
    MultiobjectiveStrategy,
)


class QehviStrategy(MultiobjectiveStrategy):
    type: Literal["QehviStrategy"] = "QehviStrategy"

    ref_point: Optional[Dict[str, float]] = None

    @validator("ref_point")
    def validate_ref_point(cls, v, values):
        """Validate that the provided refpoint matches the provided domain."""
        if v is None:
            return v
        keys = values["domain"].outputs.get_keys_by_objective(
            [MaximizeObjective, MinimizeObjective]
        )
        if sorted(keys) != sorted(v.keys()):
            raise ValueError(
                f"Provided refpoint do not match the domain, expected keys: {keys}"
            )
        return v

    @classmethod
    def is_feature_implemented(cls, my_type: Type[Feature]) -> bool:
        """Method to check if a specific feature type is implemented for the strategy

        Args:
            my_type (Type[Feature]): Feature class

        Returns:
            bool: True if the feature type is valid for the strategy chosen, False otherwise
        """
        if my_type not in [CategoricalOutput]:
            return True
        return False

    @classmethod
    def is_objective_implemented(cls, my_type: Type[Objective]) -> bool:
        """Method to check if a objective type is implemented for the strategy

        Args:
            my_type (Type[Objective]): Objective class

        Returns:
            bool: True if the objective type is valid for the strategy chosen, False otherwise
        """
        return my_type in [
            MaximizeObjective,
            MinimizeObjective,
        ]



from typing import Literal

from pydantic import Field

from bofire.data_models.kernels.api import AnyKernel, ScaleKernel
from bofire.data_models.kernels.molecular import TanimotoKernel
from bofire.data_models.priors.api import (
    BOTORCH_NOISE_PRIOR,
    BOTORCH_SCALE_PRIOR,
    AnyPrior,
)
from bofire.data_models.surrogates.botorch import BotorchSurrogate
from bofire.data_models.surrogates.scaler import ScalerEnum
from bofire.data_models.surrogates.trainable import TrainableSurrogate


class TanimotoGPSurrogate(BotorchSurrogate, TrainableSurrogate):
    type: Literal["TanimotoGPSurrogate"] = "TanimotoGPSurrogate"

    kernel: AnyKernel = Field(
        default_factory=lambda: ScaleKernel(
            base_kernel=TanimotoKernel(
                ard=True,
            ),
            outputscale_prior=BOTORCH_SCALE_PRIOR(),
        )
    )
    noise_prior: AnyPrior = Field(default_factory=lambda: BOTORCH_NOISE_PRIOR())
    scaler: ScalerEnum = ScalerEnum.IDENTITY



from itertools import chain

import numpy as np
import pytest
import torch
from botorch.acquisition.multi_objective import (  # qLogExpectedHypervolumeImprovement,; qLogNoisyExpectedHypervolumeImprovement,
    qExpectedHypervolumeImprovement,
    qNoisyExpectedHypervolumeImprovement,
)
from botorch.acquisition.multi_objective.logei import (  # qExpectedHypervolumeImprovement,
    qLogExpectedHypervolumeImprovement,
    qLogNoisyExpectedHypervolumeImprovement,
)

# qNoisyExpectedHypervolumeImprovement,
from botorch.acquisition.multi_objective.objective import GenericMCMultiOutputObjective

import bofire.data_models.acquisition_functions.api as acquisitions
import bofire.data_models.strategies.api as data_models
import bofire.strategies.api as strategies
from bofire.benchmarks.multi import C2DTLZ2, DTLZ2
from bofire.data_models.features.api import ContinuousOutput
from bofire.data_models.strategies.api import (
    PolytopeSampler as PolytopeSamplerDataModel,
)
from bofire.strategies.api import PolytopeSampler
from tests.bofire.utils.test_multiobjective import (
    dfs,
    invalid_domains,
    valid_constrained_domains,
    valid_domains,
)


@pytest.mark.parametrize(
    "domain, ref_point",
    [
        (invalid_domains[0], None),
        (invalid_domains[1], None),
        (valid_domains[0], [0]),
        (valid_domains[0], {}),
        (valid_domains[0], {"of1": 0.0, "of2": 0, "of3": 0}),
        (valid_domains[0], {"of1": 0.0}),
        (valid_domains[0], {"of1": 0.0, "of3": 0.0}),
    ],
)
def test_invalid_mobo(domain, ref_point):
    with pytest.raises(ValueError):
        data_models.MoboStrategy(domain=domain, ref_point=ref_point)


@pytest.mark.parametrize("domain", valid_constrained_domains)
def test_qnehvi_valid_constrained_objectives(domain):
    data_models.MoboStrategy(domain=domain)


@pytest.mark.parametrize(
    "domain, ref_point, experiments, expected",
    [
        (valid_domains[0], {"of1": 0.5, "of2": 10.0}, dfs[0], [0.5, -10.0]),
        (valid_domains[1], {"of1": 0.5, "of3": 0.5}, dfs[1], [0.5, 0.5]),
        (valid_domains[0], None, dfs[0], [1.0, -5.0]),
        (valid_domains[1], None, dfs[1], [1.0, 2.0]),
    ],
)
def test_mobo_get_adjusted_refpoint(domain, ref_point, experiments, expected):
    data_model = data_models.MoboStrategy(domain=domain, ref_point=ref_point)
    strategy = strategies.map(data_model)
    # hack for the test to prevent training of the model when using tell
    strategy.set_experiments(experiments)
    adjusted_ref_point = strategy.get_adjusted_refpoint()
    assert isinstance(adjusted_ref_point, list)
    assert np.allclose(expected, np.asarray(adjusted_ref_point))


@pytest.mark.parametrize(
    "strategy, use_ref_point, acqf",
    [
        (data_models.MoboStrategy, use_ref_point, acqf)
        for use_ref_point in [True, False]
        for acqf in [
            acquisitions.qEHVI,
            acquisitions.qLogEHVI,
            acquisitions.qNEHVI,
            acquisitions.qLogNEHVI,
        ]
    ],
)
def test_mobo(strategy, use_ref_point, acqf):
    # generate data
    benchmark = DTLZ2(dim=6)
    random_strategy = PolytopeSampler(
        data_model=PolytopeSamplerDataModel(domain=benchmark.domain)
    )
    experiments = benchmark.f(random_strategy._ask(n=10), return_complete=True)
    # init strategy
    data_model = strategy(
        domain=benchmark.domain,
        ref_point=benchmark.ref_point if use_ref_point else None,
        acquisition_function=acqf(),
    )
    my_strategy = strategies.map(data_model)
    my_strategy.tell(experiments)

    bacqf = my_strategy._get_acqfs(2)[0]

    assert isinstance(bacqf.objective, GenericMCMultiOutputObjective)
    if isinstance(acqf, acquisitions.qEHVI):
        assert isinstance(bacqf, qExpectedHypervolumeImprovement)
    elif isinstance(acqf, acquisitions.qNEHVI):
        assert isinstance(bacqf, qNoisyExpectedHypervolumeImprovement)
    elif isinstance(acqf, acquisitions.qLogNEHVI):
        assert isinstance(bacqf, qLogNoisyExpectedHypervolumeImprovement)
    elif isinstance(acqf, acquisitions.qLogEHVI):
        assert isinstance(bacqf, qLogExpectedHypervolumeImprovement)


@pytest.mark.parametrize(
    "acqf",
    [
        acquisitions.qEHVI,
        acquisitions.qLogEHVI,
        acquisitions.qNEHVI,
        acquisitions.qLogNEHVI,
    ],
)
def test_mobo_constraints(acqf):
    benchmark = C2DTLZ2(dim=4)
    random_strategy = PolytopeSampler(
        data_model=PolytopeSamplerDataModel(domain=benchmark.domain)
    )
    experiments = benchmark.f(random_strategy._ask(n=10), return_complete=True)
    data_model = data_models.MoboStrategy(
        domain=benchmark.domain,
        ref_point={"f_0": 1.1, "f_1": 1.1},
        acquisition_function=acqf(),
    )
    my_strategy = strategies.map(data_model)
    my_strategy.tell(experiments)
    bacqf = my_strategy._get_acqfs(2)[0]
    assert isinstance(bacqf.objective, GenericMCMultiOutputObjective)
    if isinstance(acqf, acquisitions.qEHVI):
        assert isinstance(bacqf, qExpectedHypervolumeImprovement)
    elif isinstance(acqf, acquisitions.qNEHVI):
        assert isinstance(bacqf, qNoisyExpectedHypervolumeImprovement)
    elif isinstance(acqf, acquisitions.qLogNEHVI):
        assert isinstance(bacqf, qLogNoisyExpectedHypervolumeImprovement)
    elif isinstance(acqf, acquisitions.qLogEHVI):
        assert isinstance(bacqf, qLogExpectedHypervolumeImprovement)
    assert bacqf.eta == torch.tensor(1e-3)
    assert len(bacqf.constraints) == 1
    assert torch.allclose(
        bacqf.ref_point,
        torch.tensor([-1.1, -1.1], dtype=torch.double),
    )


@pytest.mark.parametrize(
    "num_experiments, num_candidates",
    [
        (num_experiments, num_candidates)
        for num_experiments in range(8, 10)
        for num_candidates in range(1, 3)
    ],
)
@pytest.mark.slow
def test_get_acqf_input(num_experiments, num_candidates):
    # generate data
    benchmark = DTLZ2(dim=6)
    random_strategy = PolytopeSampler(
        data_model=PolytopeSamplerDataModel(domain=benchmark.domain)
    )
    experiments = benchmark.f(
        random_strategy._ask(n=num_experiments), return_complete=True
    )
    data_model = data_models.MoboStrategy(domain=benchmark.domain)
    strategy = strategies.map(data_model)
    # , ref_point=ref_pointw

    strategy.tell(experiments)
    strategy.ask(candidate_count=num_candidates, add_pending=True)

    X_train, X_pending = strategy.get_acqf_input_tensors()

    _, names = strategy.domain.inputs._get_transform_info(
        specs=strategy.surrogate_specs.input_preprocessing_specs
    )

    assert torch.is_tensor(X_train)
    assert torch.is_tensor(X_pending)
    assert X_train.shape == (
        num_experiments,
        len(set(chain(*names.values()))),
    )
    assert X_pending.shape == (
        num_candidates,
        len(set(chain(*names.values()))),
    )


def test_no_objective():
    domain = DTLZ2(dim=6).domain
    experiments = DTLZ2(dim=6).f(domain.inputs.sample(10), return_complete=True)
    domain.outputs.features.append(ContinuousOutput(key="ignore", objective=None))
    experiments["ignore"] = experiments["f_0"] + 6
    experiments["valid_ignore"] = 1
    data_model = data_models.MoboStrategy(
        domain=domain, ref_point={"f_0": 1.1, "f_1": 1.1}
    )
    recommender = strategies.map(data_model=data_model)
    recommender.tell(experiments=experiments)
    candidates = recommender.ask(candidate_count=1)
    recommender.to_candidates(candidates)



from abc import abstractmethod
from typing import Literal, Union

from pydantic import Field, validator
from typing_extensions import Annotated

from bofire.data_models.base import BaseModel
from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.surrogates.api import (
    MixedSingleTaskGPSurrogate,
    SingleTaskGPSurrogate,
)


class OutlierDetection(BaseModel):
    type: str

    @property
    @abstractmethod
    def inputs(self) -> Inputs:  # type: ignore
        pass

    @property
    @abstractmethod
    def outputs(self) -> Outputs:  # type: ignore
        pass


class IterativeTrimming(OutlierDetection):
    """Remove outliers using Robust GP

    Paper: Robust Gaussian Process Regression Based on Iterative Trimming.
    https://arxiv.org/pdf/2011.11057.pdf

    Attributes:
        alpha1 (float in (0, 1)): Trimming parameter.
        alpha2 (float in (0, 1)): Reweighting parameter.
        nsh (int (>=1)): Number of shrinking iterations.
        ncc (int (>=1)): Number of concentrating iterations.
        nrw (int (>=1)): Number of reweighting iterations.
        base_gp (SingleTaskGPSurrogate): Gaussian process model for outlier detection.
    """

    type: Literal["IterativeTrimming"] = "IterativeTrimming"
    alpha1: Annotated[float, Field(gt=0.0, lt=1.0)] = 0.5
    alpha2: Annotated[float, Field(gt=0.0, lt=1.0)] = 0.975
    nsh: Annotated[int, Field(ge=1)] = 2
    ncc: Annotated[int, Field(ge=1)] = 2
    nrw: Annotated[int, Field(ge=1)] = 1
    base_gp: Union[SingleTaskGPSurrogate, MixedSingleTaskGPSurrogate]

    @validator("base_gp")
    def validate_base_gp(cls, v, values):
        # validate that all base_gps are single output surrogates
        # TODO: this restriction has to be removed at some point

        if len(v.outputs) != 1:
            raise ValueError("Only single output base_gps allowed.")

        return v

    @property
    def inputs(self) -> Inputs:
        return self.base_gp.inputs

    @property
    def outputs(self) -> Outputs:
        return self.base_gp.outputs



from bofire.data_models.domain.constraints import Constraints  # noqa: F401
from bofire.data_models.domain.domain import Domain  # noqa: F401
from bofire.data_models.domain.features import Features, Inputs, Outputs  # noqa: F401



import random

import pandas as pd
import pytest

import bofire.data_models.strategies.api as data_models
import bofire.strategies.api as strategies
from bofire.benchmarks.multi import DTLZ2
from bofire.benchmarks.single import Ackley
from bofire.data_models.strategies.api import (
    PolytopeSampler as PolytopeSamplerDataModel,
)
from bofire.strategies.api import PolytopeSampler
from tests.bofire.strategies.test_qehvi import VALID_BOTORCH_QEHVI_STRATEGY_SPEC
from tests.bofire.strategies.test_sobo import VALID_BOTORCH_SOBO_STRATEGY_SPEC

STRATEGY_SPECS_SINGLE_OBJECTIVE = {
    # BoTorchSoboAdditiveStrategy: VALID_BOTORCH_SOBO_STRATEGY_SPEC,
    data_models.MultiplicativeSoboStrategy: VALID_BOTORCH_SOBO_STRATEGY_SPEC,
}
STRATEGY_SPECS_MULTI_OBJECTIVE = {
    data_models.QehviStrategy: VALID_BOTORCH_QEHVI_STRATEGY_SPEC,
    data_models.QnehviStrategy: VALID_BOTORCH_QEHVI_STRATEGY_SPEC,
}


# TODO: check this properly
@pytest.mark.parametrize(
    "cls, spec, categorical, descriptor, candidate_count",
    [
        (cls, specs, categorical, descriptor, random.randint(1, 2))
        for cls, specs in STRATEGY_SPECS_SINGLE_OBJECTIVE.items()
        for categorical in [True, False]
        for descriptor in [True, False]
    ],
)
@pytest.mark.slow
def test_ask_single_objective(cls, spec, categorical, descriptor, candidate_count):
    # generate data
    benchmark = Ackley(categorical=categorical, descriptor=descriptor)
    random_strategy = PolytopeSampler(
        data_model=PolytopeSamplerDataModel(domain=benchmark.domain)
    )
    experiments = benchmark.f(random_strategy.ask(10), return_complete=True)

    # set up of the strategy
    data_model = cls(**{**spec, "domain": benchmark.domain})
    print("data_model:", type(data_model))
    strategy = strategies.map(data_model)
    strategy.tell(experiments)

    # ask
    candidates = strategy.ask(candidate_count=candidate_count)

    assert isinstance(candidates, pd.DataFrame)
    assert len(candidates) == candidate_count

    # TODO: add check of convergence towards the optimum (0)


@pytest.mark.parametrize(
    "cls, spec, use_ref_point, candidate_count",
    [
        (cls, specs, use_ref_point, random.randint(1, 2))
        for cls, specs in STRATEGY_SPECS_MULTI_OBJECTIVE.items()
        for use_ref_point in [True, False]
        # for categorical in [True, False]
        # for descriptor in [True, False]
    ],
)
@pytest.mark.slow  # use pytest . --runslow in command line to include these tests
def test_ask_multi_objective(cls, spec, use_ref_point, candidate_count):
    # generate data
    benchmark = DTLZ2(
        dim=6
    )  # TODO: expand benchmark also towards categorical features?
    random_strategy = PolytopeSampler(
        data_model=PolytopeSamplerDataModel(domain=benchmark.domain)
    )
    experiments = benchmark.f(random_strategy.ask(10), return_complete=True)

    # set up of the strategy
    data_model = cls(
        **{**spec, "domain": benchmark.domain},
        # domain=benchmark.domain,
        ref_point=benchmark.ref_point if use_ref_point else None
    )
    strategy = strategies.map(data_model=data_model)
    strategy.tell(experiments)

    # ask
    candidates = strategy.ask(candidate_count=candidate_count)

    assert isinstance(candidates, pd.DataFrame)
    assert len(candidates) == candidate_count

    # TODO: add check of convergence towards the optimum (0)



from typing import Literal, Optional, Type

from pydantic import Field, validator

from bofire.data_models.acquisition_functions.api import (
    AnySingleObjectiveAcquisitionFunction,
    qLogNEI,
)
from bofire.data_models.features.api import CategoricalOutput, Feature
from bofire.data_models.objectives.api import ConstrainedObjective, Objective
from bofire.data_models.strategies.predictives.botorch import BotorchStrategy


class SoboBaseStrategy(BotorchStrategy):
    acquisition_function: AnySingleObjectiveAcquisitionFunction = Field(
        default_factory=lambda: qLogNEI()
    )

    @classmethod
    def is_feature_implemented(cls, my_type: Type[Feature]) -> bool:
        """Method to check if a specific feature type is implemented for the strategy

        Args:
            my_type (Type[Feature]): Feature class

        Returns:
            bool: True if the feature type is valid for the strategy chosen, False otherwise
        """
        if my_type not in [CategoricalOutput]:
            return True
        return False

    @classmethod
    def is_objective_implemented(cls, my_type: Type[Objective]) -> bool:
        """Method to check if a objective type is implemented for the strategy

        Args:
            my_type (Type[Objective]): Objective class

        Returns:
            bool: True if the objective type is valid for the strategy chosen, False otherwise
        """
        return True


class SoboStrategy(SoboBaseStrategy):
    type: Literal["SoboStrategy"] = "SoboStrategy"

    @validator("domain")
    def validate_is_singleobjective(cls, v, values):
        if len(v.outputs) == 1:
            return v
        if (
            len(v.outputs.get_by_objective(excludes=ConstrainedObjective))
            - len(v.outputs.get_by_objective(includes=None, excludes=Objective))
        ) > 1:
            raise ValueError(
                "SOBO strategy can only deal with one no-constraint objective."
            )
        return v


class AdditiveSoboStrategy(SoboBaseStrategy):
    type: Literal["AdditiveSoboStrategy"] = "AdditiveSoboStrategy"
    use_output_constraints: bool = True

    @validator("domain")
    def validate_is_multiobjective(cls, v, values):
        if (len(v.outputs.get_by_objective(Objective))) < 2:
            raise ValueError(
                "Additive SOBO strategy requires at least 2 outputs with objectives. Consider SOBO strategy instead."
            )
        return v


class MultiplicativeSoboStrategy(SoboBaseStrategy):
    type: Literal["MultiplicativeSoboStrategy"] = "MultiplicativeSoboStrategy"

    @validator("domain")
    def validate_is_multiobjective(cls, v, values):
        if (len(v.outputs.get_by_objective(Objective))) < 2:
            raise ValueError(
                "Multiplicative SOBO strategy requires at least 2 outputs with objectives. Consider SOBO strategy instead."
            )
        return v


class CustomSoboStrategy(SoboBaseStrategy):
    type: Literal["CustomSoboStrategy"] = "CustomSoboStrategy"
    use_output_constraints: bool = True
    dump: Optional[str] = None



import importlib

import pytest
from pandas.testing import assert_frame_equal

import bofire.surrogates.api as surrogates
from bofire.benchmarks.single import Himmelblau
from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.enum import CategoricalEncodingEnum
from bofire.data_models.features.api import (
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
)
from bofire.data_models.surrogates.api import XGBoostSurrogate

XGB_AVAILABLE = importlib.util.find_spec("xgboost") is not None


@pytest.mark.skipif(not XGB_AVAILABLE, reason="requires xgboost")
def test_XGBoostSurrogate():
    benchmark = Himmelblau()
    samples = benchmark.domain.inputs.sample(10)
    experiments = benchmark.f(samples, return_complete=True)
    #
    data_model = XGBoostSurrogate(
        inputs=benchmark.domain.inputs, outputs=benchmark.domain.outputs, n_estimators=2
    )
    surrogate = surrogates.map(data_model)
    assert isinstance(surrogate, surrogates.XGBoostSurrogate)
    assert surrogate.input_preprocessing_specs == {}
    assert surrogate.is_fitted is False
    # fit it
    surrogate.fit(experiments=experiments)
    assert surrogate.is_fitted is True
    # predict it
    preds = surrogate.predict(experiments)
    # dump it
    dump = surrogate.dumps()
    # load it
    surrogate2 = surrogates.map(data_model)
    surrogate2.loads(dump)
    preds2 = surrogate2.predict(experiments)
    assert_frame_equal(preds, preds2)
    assert_frame_equal(preds, preds2)


def test_XGBoostSurrogate_categorical():
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
        + [CategoricalInput(key="x_cat", categories=["mama", "papa"])]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments = inputs.sample(n=10)
    experiments.eval("y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments.loc[experiments.x_cat == "mama", "y"] *= 5.0
    experiments.loc[experiments.x_cat == "papa", "y"] /= 2.0
    experiments["valid_y"] = 1
    data_model = XGBoostSurrogate(inputs=inputs, outputs=outputs, n_estimators=2)
    assert data_model.input_preprocessing_specs == {
        "x_cat": CategoricalEncodingEnum.ONE_HOT
    }
    surrogate = surrogates.map(data_model)
    surrogate.fit(experiments)



from typing import Any, Dict, List

import pandas as pd
import plotly.graph_objs as go


def compose_annotation(
    caption: str, x: float = 0.0, y: float = -0.15
) -> List[Dict[str, Any]]:
    if not caption:
        return []
    return [
        {
            "showarrow": False,
            "text": caption,
            "x": x,
            "xanchor": "left",
            "xref": "paper",
            "y": y,
            "yanchor": "top",
            "yref": "paper",
            "align": "left",
        },
    ]


def plot_feature_importance_by_feature_plotly(
    sensitivity_values: Dict[str, pd.DataFrame],
    relative: bool = False,
    show_std: bool = False,
    caption: str = "",
    importance_measure: str = "",
) -> go.Figure:
    """One plot per metric, showing importances by feature.

    Main part of this code is taken from Ax (https://github.com/facebook/Ax/blob/main/ax/plot/feature_importances.py).

    Args:
        sensitivity_values (Dict[str, pd.DataFrame]): The sensitivity values for each metric in a dict format.
            It takes the following format if only the sensitivity value is plotted:
            `{"metric1":pd.DataFrame}` where the columns of the dataframe are the
            features and the index is called `mean` and `std`.
        relative (bool, optional): Whether to normalize feature importances so that they add to 1. Defaults to False.
        show_std (bool, optional): Whether to show the standard deviation in the plot. Defaults to False.
        caption: An HTML-formatted string to place at the bottom of the plot.
        importance_measure: The name of the importance metric to be added to the title.
    Returns:
        go.Figure: Figure of feature importances.
    """
    traces = []
    dropdown = []
    for i, metric_name in enumerate(sorted(sensitivity_values.keys())):
        importances = sensitivity_values[metric_name]
        df = pd.DataFrame(
            [
                {"Feature": feature, "Importance": importances.loc["mean", feature]}
                for feature in importances.columns
            ]
        )
        if show_std and "std" in importances.index:
            error_x = {
                "type": "data",
                "array": [
                    importances.loc["std", feature] for feature in importances.columns
                ],
                "visible": True,
            }
        else:
            error_x = None
        if relative:
            df["Importance"] = df["Importance"].div(df["Importance"].sum())
        traces.append(
            go.Bar(
                name="Importance",
                orientation="h",
                visible=i == 0,
                x=df["Importance"],
                y=df["Feature"],
                error_x=error_x,
                opacity=0.8,
            )
        )

        is_visible = [False] * len(sensitivity_values)
        is_visible[i] = True
        dropdown.append(
            {"args": ["visible", is_visible], "label": metric_name, "method": "restyle"}
        )
    if not traces:
        raise NotImplementedError("No traces found for metric")

    updatemenus = [
        {
            "x": 0,
            "y": 1,
            "yanchor": "top",
            "xanchor": "left",
            "buttons": dropdown,
            "pad": {
                "t": -40
            },  # hack to put dropdown below title regardless of number of features
        }
    ]
    features = traces[0].y
    title = (
        "Relative Feature Importances" if relative else "Absolute Feature Importances"
    )
    if importance_measure:
        title = title + " based on " + importance_measure
    layout = go.Layout(
        height=200 + len(features) * 20,
        hovermode="closest",
        margin=go.layout.Margin(
            l=8 * min(max(len(idx) for idx in features), 75)
        ),  # noqa E741
        showlegend=False,
        title=title,
        updatemenus=updatemenus,
        annotations=compose_annotation(caption=caption),
    )

    if relative:
        layout.update({"xaxis": {"tickformat": ".0%"}})

    return go.Figure(data=traces, layout=layout)



"""
Module for test_kernels that operate on fingerprint representations (bit vectors or count vectors).
Author: Ryan-Rhys Griffiths and Austin Tripp 2022
"""
# This code was copied from GAUCHE: https://github.com/leojklarner/gauche/blob/main/gauche/kernels/fingerprint_kernels/base_fingerprint_kernel.py

import torch
from gpytorch.kernels import Kernel


def default_postprocess_script(x):
    return x


def batch_tanimoto_sim(
    x1: torch.Tensor, x2: torch.Tensor, eps: float = 1e-6
) -> torch.Tensor:
    """
    Tanimoto between two batched tensors, across last 2 dimensions.
    eps argument ensures numerical stability if all zero tensors are added.
    """
    # Tanimoto distance is proportional to (<x, y>) / (||x||^2 + ||y||^2 - <x, y>) where x and y are bit vectors
    assert x1.ndim >= 2 and x2.ndim >= 2
    dot_prod = torch.matmul(x1, torch.transpose(x2, -1, -2))
    # x1_sum = torch.sum(x1**2, dim=-1, keepdims=True)
    # x2_sum = torch.sum(x2**2, dim=-1, keepdims=True)
    x1_sum = torch.sum(x1**2, dim=-1).unsqueeze(-1)
    x2_sum = torch.sum(x2**2, dim=-1).unsqueeze(-1)
    return (dot_prod + eps) / (
        eps + x1_sum + torch.transpose(x2_sum, -1, -2) - dot_prod
    )


class BitDistance(torch.nn.Module):
    """
    Distance module for bit vector test_kernels.
    """

    def __init__(self, postprocess_script=default_postprocess_script):
        super().__init__()
        self._postprocess = postprocess_script

    def _sim(self, x1, x2, postprocess, x1_eq_x2=False, metric="tanimoto"):
        """
        Computes the similarity between x1 and x2

        Args:
            x1 (Tensor): First set of data where b is a batch dimension. Has shape `n x d` or `b x n x d`
            x2 (Tensor): Second set of data where b is a batch dimension. Has shape `m x d` or `b x m x d`
            postprocess (bool): Whether to apply a postprocess script
            x1_eq_x2 (bool, optional): Is x1 equal to x2. Defaults to False
            metric (str): String specifying the similarity metric. One of ['tanimoto']. Defaults to 'tanimoto'

        Raises:
            RuntimeError: If tanimoto is not used as the similarity metric.

        Returns:
            Tensor: corresponding to the similarity matrix between `x1` and `x2`
        """
        # Branch for Tanimoto metric
        if metric == "tanimoto":
            res = batch_tanimoto_sim(x1, x2)
            res.clamp_min_(0)  # zero out negative values
            return self._postprocess(res) if postprocess else res
        else:
            raise RuntimeError(
                "Similarity metric not supported. Available options are 'tanimoto'"
            )


class BitKernel(Kernel):
    """
     Base class for test_kernels that operate on bit or count vectors such as ECFP fingerprints or RDKit fragments.
     In the typical use case, test_kernels inheriting from this class will specify a similarity metric such as Tanimoto,
     MinMax etc. This kernel does not have an `outputscale` parameter. To add a scaling parameter, decorate this kernel
     with a `gpytorch.test_kernels.ScaleKernel`. This base :class:`BitKernel` class does not include a lengthscale
     parameter `Theta`, in contrast to many common kernel functions.

    Args:
        metric (str): The similarity metric to use. One of ['tanimoto']. Defaults to ''
    """

    def __init__(self, metric="", **kwargs):
        super().__init__(**kwargs)
        self.metric = metric

    def forward(self, x1, x2, **params):
        return self.covar_dist(x1, x2, **params)

    def covar_dist(
        self,
        x1,
        x2,
        last_dim_is_batch=False,
        dist_postprocess_func=default_postprocess_script,
        postprocess=True,
        **params,
    ):
        """
        This is a helper method for computing the bit vector similarity between
        all pairs of points in x1 and x2.

        Args:
            x1 (Tensor): First set of data. Has shape `n x d` or `b1 x ... x bk x n x d`
            x2 (Tensor): Second set of data. Has shape `m x d` or `b1 x ... x bk x m x d`
            last_dim_is_batch (bool, optional): Is the last dimension of the data a batch dimension or not?.
            Defaults to False
            postprocess (bool): Whether to apply a postprocess script

        Returns:
            Tensor: corresponding to the distance matrix between `x1` and `x2`.
            The shape depends on the kernel's mode
            * `diag=False`
            * `diag=False` and `last_dim_is_batch=True`: (`b x d x n x n`)
            * `diag=True`
            * `diag=True` and `last_dim_is_batch=True`: (`b x d x n`)
        """
        if last_dim_is_batch:
            x1 = x1.transpose(-1, -2).unsqueeze(-1)
            x2 = x2.transpose(-1, -2).unsqueeze(-1)

        x1_eq_x2 = torch.equal(x1, x2)

        # torch scripts expect tensors
        postprocess = torch.tensor(postprocess)

        res = None

        # Cache the Distance object or else JIT will recompile every time
        if (
            not self.distance_module
            or self.distance_module._postprocess != dist_postprocess_func
        ):
            self.distance_module = BitDistance(dist_postprocess_func)

        res = self.distance_module._sim(x1, x2, postprocess, x1_eq_x2, self.metric)

        return res



import pytest
from pandas.testing import assert_frame_equal

import bofire.surrogates.api as surrogates
from bofire.benchmarks.single import Himmelblau
from bofire.data_models.surrogates.api import SaasSingleTaskGPSurrogate


def test_SaasSingleTaskGPSurrogate_invalid_thinning():
    bench = Himmelblau()
    with pytest.raises(ValueError):
        SaasSingleTaskGPSurrogate(
            inputs=bench.domain.inputs,
            outputs=bench.domain.outputs,
            warmup_steps=32,
            num_samples=16,
            thinning=18,
        )


def test_SaasSingleTaskGPSurrogate():
    bench = Himmelblau()
    samples = bench.domain.inputs.sample(10)
    experiments = bench.f(samples, return_complete=True)
    data_model = SaasSingleTaskGPSurrogate(
        inputs=bench.domain.inputs,
        outputs=bench.domain.outputs,
        warmup_steps=32,
        num_samples=16,
        thinning=4,
    )
    gp = surrogates.map(data_model)
    gp.fit(experiments=experiments)
    dump = gp.dumps()
    gp2 = surrogates.map(data_model=data_model)
    gp2.loads(dump)
    preds = gp.predict(experiments)
    assert preds.shape == (10, 2)
    preds2 = gp.predict(experiments)
    assert_frame_equal(preds, preds2)



import pytest

from bofire.benchmarks.api import Himmelblau, Hyperopt
from bofire.data_models.enum import RegressionMetricsEnum
from bofire.data_models.surrogates.api import SingleTaskGPSurrogate

benchmark = Himmelblau()
experiments = benchmark.f(benchmark.domain.inputs.sample(9), return_complete=True)


def test_Hyperopt_invalid():
    surrogate_data = SingleTaskGPSurrogate(
        inputs=benchmark.domain.inputs,
        outputs=benchmark.domain.outputs,
        hyperconfig=None,
    )
    with pytest.raises(ValueError, match="No hyperoptimization configuration found."):
        Hyperopt(surrogate_data=surrogate_data, training_data=experiments, folds=5)


@pytest.mark.parametrize("n_candidates", [1, 2])
def test_Hyperopt(n_candidates: int):
    surrogate_data = SingleTaskGPSurrogate(
        inputs=benchmark.domain.inputs,
        outputs=benchmark.domain.outputs,
    )
    hy = Hyperopt(surrogate_data=surrogate_data, training_data=experiments, folds=3)
    assert hy.target_metric == surrogate_data.hyperconfig.target_metric
    assert hy.domain == surrogate_data.hyperconfig.domain
    candidates = hy.domain.inputs.sample(n_candidates)
    results = hy.f(candidates=candidates)
    assert len(results) == len(candidates)
    assert f"valid_{hy.target_metric.name}" in results.columns
    for e in RegressionMetricsEnum:
        assert e.name in results.columns
    assert len(results.columns) == 8



import importlib

import pytest

import bofire.data_models.molfeatures.names as names

RDKIT_AVAILABLE = importlib.util.find_spec("rdkit") is not None


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_fragments():
    from rdkit.Chem import Descriptors

    fragment_list = [
        item[0] for item in Descriptors.descList if item[0].startswith("fr_")
    ]

    assert names.fragments == fragment_list


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_mordred():
    from mordred import Calculator
    from mordred import descriptors as mordred_descriptors

    calc = Calculator(mordred_descriptors, ignore_3D=False)
    assert names.mordred == [str(d) for d in calc.descriptors]



import itertools
from math import nan

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal
from pydantic.error_wrappers import ValidationError
from pytest import fixture

from bofire.data_models.api import Outputs
from bofire.data_models.base import BaseModel
from bofire.data_models.constraints.api import (
    LinearEqualityConstraint,
    NChooseKConstraint,
)
from bofire.data_models.domain.api import Domain
from bofire.data_models.domain.constraints import Constraints
from bofire.data_models.domain.features import Inputs
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
    DiscreteInput,
    Feature,
    Input,
    Output,
)
from bofire.data_models.objectives.api import (
    ConstrainedObjective,
    MaximizeObjective,
    MaximizeSigmoidObjective,
    Objective,
    TargetObjective,
)
from bofire.utils.subdomain import get_subdomain


def test_empty_domain():
    assert Domain() == Domain(
        inputs=Inputs(), outputs=Outputs(), constraints=Constraints()
    )


class Bla(BaseModel):
    a: int


nf = Bla(a=1)

obj = TargetObjective(target_value=1, steepness=2, tolerance=3, w=0.5)

if_ = CategoricalInput(key="if", categories=["a", "b"])
of_ = ContinuousOutput(key="of", objective=obj)
if1 = ContinuousInput(key="f1", bounds=(0, 10))
if1_ = ContinuousInput(key="f1", bounds=(0, 1))
if2 = ContinuousInput(key="f2", bounds=(0, 10))
of1 = ContinuousOutput(key="f1", objective=obj)
of1_ = ContinuousOutput(key="f1", objective=obj)
of2 = ContinuousOutput(key="f2", objective=obj)
of3 = ContinuousOutput(key="of3", objective=obj)


@fixture
def input_list():
    return [
        ContinuousInput(key="ci1", bounds=(0, 1)),
        ContinuousInput(key="ci2", bounds=(0, 1)),
    ]


@fixture
def output_list():
    objective = TargetObjective(target_value=1, steepness=2, tolerance=3, w=0.5)
    return [
        ContinuousOutput(key="co1", objective=objective),
        ContinuousOutput(key="co2", objective=objective),
    ]


@fixture
def constraint_list(input_list):
    return [
        LinearEqualityConstraint(
            features=[inp.key for inp in input_list],
            coefficients=[1.0] * len(input_list),
            rhs=11,
        )
    ]


def test_from_lists(input_list, output_list, constraint_list):
    assert Domain.from_lists(
        inputs=input_list, outputs=output_list, constraints=constraint_list
    ) == Domain(
        inputs=Inputs(features=input_list),
        outputs=Outputs(features=output_list),
        constraints=Constraints(constraints=constraint_list),
    )


@pytest.mark.parametrize(
    "inputs, outputs",
    [
        ([if1, if1_], []),
        ([], [of1, of1_]),
        ([if1], [of1]),
    ],
)
def test_duplicate_feature_names(inputs, outputs):
    with pytest.raises(ValidationError):
        Domain(
            inputs=inputs,
            outputs=outputs,
        )


@pytest.mark.parametrize(
    "inputs, outputs, constraints",
    [
        # input features
        ([nf], [], []),
        # output features
        ([], [nf], []),
        # constraints
        # ([], [], [nf]),
    ],
)
def test_invalid_domain_arg_types(inputs, outputs, constraints):
    with pytest.raises(Exception):
        Domain(
            inputs=inputs,
            outputs=outputs,
            constraints=constraints,
        )


@pytest.mark.parametrize(
    "inputs, constraints",
    [
        (
            [
                ContinuousInput(key="if1", bounds=(0, 1)),
                DiscreteInput(key="if2", values=[0.2, 0.7, 1.0]),
            ],
            [
                LinearEqualityConstraint(
                    features=["if1", "if2"], coefficients=[1.0, 1.0], rhs=11
                )
            ],
        )
    ],
)
def test_invalid_type_in_linear_constraints(inputs, constraints):
    with pytest.raises(ValidationError):
        Domain(inputs=inputs, constraints=constraints)


@pytest.mark.parametrize(
    "inputs, outputs, constraints",
    [
        (
            [if1, if2],
            [of3],
            [
                NChooseKConstraint(
                    features=["f1", "f2"],
                    min_count=0,
                    max_count=1,
                    none_also_valid=True,
                )
            ],
        ),
        (
            [if1, if2],
            [],
            [
                NChooseKConstraint(
                    features=["f1", "f2"],
                    min_count=0,
                    max_count=2,
                    none_also_valid=True,
                )
            ],
        ),
    ],
)
def test_valid_constraints_in_domain(outputs, inputs, constraints):
    Domain(
        inputs=inputs,
        outputs=outputs,
        constraints=constraints,
    )


@pytest.mark.parametrize(
    "inputs, outputs, constraints",
    [
        (
            [],
            [],
            [
                NChooseKConstraint(
                    features=["x", "t55"],
                    min_count=0,
                    max_count=1,
                    none_also_valid=True,
                )
            ],
        ),
        (
            [if1],
            [],
            [
                NChooseKConstraint(
                    features=["f1", "x"],
                    min_count=0,
                    max_count=2,
                    none_also_valid=True,
                )
            ],
        ),
        (
            [],
            [of1],
            [
                NChooseKConstraint(
                    features=["f1", "x"],
                    min_count=0,
                    max_count=1,
                    none_also_valid=True,
                )
            ],
        ),
        (
            [if1],
            [of1],
            [
                NChooseKConstraint(
                    features=["f1", "f2"],
                    min_count=0,
                    max_count=1,
                    none_also_valid=True,
                )
            ],
        ),
    ],
)
def test_unknown_features_in_domain(outputs, inputs, constraints):
    with pytest.raises(ValidationError):
        Domain(
            inputs=inputs,
            outputs=outputs,
            constraints=constraints,
        )


@pytest.mark.parametrize(
    "domain, data",
    [
        (
            Domain(),
            [],
        ),
        (
            Domain(
                inputs=Inputs(
                    features=[
                        CategoricalInput(
                            key="f1",
                            categories=["c11", "c12"],
                        ),
                    ]
                ),
            ),
            [
                [("f1", "c11"), ("f1", "c12")],
            ],
        ),
        (
            Domain(
                inputs=Inputs(
                    features=[
                        CategoricalInput(
                            key="f1",
                            categories=["c11", "c12"],
                        ),
                        CategoricalInput(
                            key="f2",
                            categories=["c21", "c22", "c23"],
                        ),
                        CategoricalInput(
                            key="f3",
                            categories=["c31", "c32"],
                        ),
                    ]
                ),
            ),
            [
                [("f1", "c11"), ("f1", "c12")],
                [("f2", "c21"), ("f2", "c22"), ("f2", "c23")],
                [("f3", "c31"), ("f3", "c32")],
            ],
        ),
    ],
)
def test_categorical_combinations_of_domain_defaults(domain, data):
    expected = list(itertools.product(*data))
    assert domain.inputs.get_categorical_combinations() == expected


@pytest.mark.parametrize(
    "domain, data, include, exclude",
    [
        (
            Domain(
                inputs=Inputs(
                    features=[
                        CategoricalInput(
                            key="f1",
                            categories=["c11", "c12"],
                        ),
                        CategoricalDescriptorInput(
                            key="f2",
                            categories=["c21", "c22"],
                            descriptors=["d21", "d22"],
                            values=[[1, 2], [3, 4]],
                        ),
                    ]
                ),
            ),
            [
                [("f1", "c11"), ("f1", "c12")],
            ],
            CategoricalInput,
            CategoricalDescriptorInput,
        ),
        (
            Domain(
                inputs=Inputs(
                    features=[
                        CategoricalInput(
                            key="f1",
                            categories=["c11", "c12"],
                        ),
                        CategoricalDescriptorInput(
                            key="f2",
                            categories=["c21", "c22"],
                            descriptors=["d21", "d22"],
                            values=[[1, 2], [3, 4]],
                        ),
                    ]
                ),
            ),
            [
                [("f2", "c21"), ("f2", "c22")],
                [("f1", "c11"), ("f1", "c12")],
            ],
            CategoricalInput,
            None,
        ),
        (
            Domain(
                inputs=Inputs(
                    features=[
                        CategoricalInput(
                            key="f1",
                            categories=["c11", "c12"],
                        ),
                        CategoricalDescriptorInput(
                            key="f2",
                            categories=["c21", "c22"],
                            descriptors=["d21", "d22"],
                            values=[[1, 2], [3, 4]],
                        ),
                    ]
                ),
            ),
            [
                [("f2", "c21"), ("f2", "c22")],
            ],
            CategoricalDescriptorInput,
            None,
        ),
        (
            Domain(
                inputs=Inputs(
                    features=[
                        CategoricalInput(
                            key="f1",
                            categories=["c11", "c12"],
                        ),
                        CategoricalDescriptorInput(
                            key="f2",
                            categories=["c21", "c22"],
                            descriptors=["d21", "d22"],
                            values=[[1, 2], [3, 4]],
                        ),
                    ]
                ),
            ),
            [],
            CategoricalDescriptorInput,
            CategoricalInput,
        ),
    ],
)
def test_categorical_combinations_of_domain_filtered(domain, data, include, exclude):
    expected = list(itertools.product(*data))
    assert (
        domain.inputs.get_categorical_combinations(include=include, exclude=exclude)
        == expected
    )


data = pd.DataFrame.from_dict(
    {
        "x1": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
        "x2": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
        "out1": [nan, 1.0, 2.0, 3.0, nan, nan],
        "out2": [nan, 1.0, 2.0, 3.0, 4.0, 5.0],
        "valid_out1": [1, 0, 1, 1, 1, 1],
        "valid_out2": [1, 1, 0, 1, 1, 0],
    }
)

if1 = ContinuousInput(key="x1", bounds=(1, 10))
if2 = ContinuousInput(key="x2", bounds=(1, 10))

of1 = ContinuousOutput(key="out1", objective=obj)
of2 = ContinuousOutput(key="out2", objective=obj)

c1 = LinearEqualityConstraint(features=["x1", "x2"], coefficients=[5, 5], rhs=15.0)

of1_ = ContinuousOutput(key="out3", objective=None)
of2_ = ContinuousOutput(key="out4", objective=None)

domain = Domain(
    inputs=Inputs(features=[if1, if2]), outputs=Outputs(features=[of1, of2])
)
domain2 = Domain(
    inputs=Inputs(features=[if1, if2]),
    outputs=Outputs(features=[of1, of2]),
    constraints=Constraints(constraints=[c1]),
)


@pytest.mark.parametrize("domain", [domain, domain2])
def test_domain_serialize(domain):
    print("domain:", domain)
    import json

    print("dict:", json.dumps(domain.dict(), indent=4))
    ndomain = Domain(**json.loads(domain.json()))
    assert ndomain == domain


@pytest.mark.parametrize(
    "domain, data, output_feature_keys, expected",
    [
        (
            domain,
            data,
            None,
            pd.DataFrame.from_dict(
                {
                    "x1": [4.0],
                    "x2": [4.0],
                    "out1": [3.0],
                    "out2": [3.0],
                    "valid_out1": [1],
                    "valid_out2": [1],
                }
            ),
        ),
        (
            domain,
            data,
            [],
            pd.DataFrame.from_dict(
                {
                    "x1": [4.0],
                    "x2": [4.0],
                    "out1": [3.0],
                    "out2": [3.0],
                    "valid_out1": [1],
                    "valid_out2": [1],
                }
            ),
        ),
        (
            domain,
            data,
            ["out1", "out2"],
            pd.DataFrame.from_dict(
                {
                    "x1": [4.0],
                    "x2": [4.0],
                    "out1": [3.0],
                    "out2": [3.0],
                    "valid_out1": [1],
                    "valid_out2": [1],
                }
            ),
        ),
        (
            domain,
            data,
            ["out2"],
            pd.DataFrame.from_dict(
                {
                    "x1": [2.0, 4.0, 5.0],
                    "x2": [2.0, 4.0, 5.0],
                    "out1": [1.0, 3.0, nan],
                    "out2": [1.0, 3.0, 4.0],
                    "valid_out1": [0, 1, 1],
                    "valid_out2": [1, 1, 1],
                }
            ),
        ),
    ],
)
def test_preprocess_experiments_all_valid_outputs(
    domain, data, output_feature_keys, expected
):
    experiments = domain.outputs.preprocess_experiments_all_valid_outputs(
        data, output_feature_keys
    )
    assert_frame_equal(experiments.reset_index(drop=True), expected, check_dtype=False)


@pytest.mark.parametrize(
    "domain, data, expected",
    [
        (
            domain,
            data,
            pd.DataFrame.from_dict(
                {
                    "x1": [2, 3, 4, 5],
                    "x2": [2, 3, 4, 5],
                    "out1": [1, 2, 3, nan],
                    "out2": [1, 2, 3, 4],
                    "valid_out1": [0, 1, 1, 1],
                    "valid_out2": [1, 0, 1, 1],
                }
            ),
        )
    ],
)
def test_preprocess_experiments_any_valid_output(domain, data, expected):
    experiments = domain.outputs.preprocess_experiments_any_valid_output(data)
    assert experiments["x1"].tolist() == expected["x1"].tolist()
    assert experiments["out2"].tolist() == expected["out2"].tolist()


@pytest.mark.parametrize(
    "domain, data, expected",
    [
        (
            domain,
            data,
            pd.DataFrame.from_dict(
                {
                    "x1": [2, 4, 5],
                    "x2": [2, 4, 5],
                    "out1": [1, 3, nan],
                    "out2": [1, 3, 4],
                    "valid_out1": [0, 1, 1],
                    "valid_out2": [1, 1, 1],
                }
            ),
        )
    ],
)
def test_preprocess_experiments_one_valid_output(domain, data, expected):
    experiments = domain.outputs.preprocess_experiments_one_valid_output("out2", data)
    assert experiments["x1"].tolist() == expected["x1"].tolist()
    assert np.isnan(experiments["out1"].tolist()[2])
    assert experiments["out2"].tolist() == expected["out2"].tolist()


def test_coerce_invalids():
    domain = Domain(
        inputs=Inputs(features=[if1, if2]), outputs=Outputs(features=[of1, of2])
    )
    experiments = domain.coerce_invalids(data)
    expected = pd.DataFrame.from_dict(
        {
            "x1": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
            "x2": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
            "out1": [nan, nan, 2.0, 3.0, nan, nan],
            "out2": [nan, 1.0, nan, 3.0, 4.0, nan],
            "valid_out1": [1, 0, 1, 1, 1, 1],
            "valid_out2": [1, 1, 0, 1, 1, 0],
        }
    )
    assert_frame_equal(experiments, expected, check_dtype=False)


@pytest.mark.parametrize("method", ["mean", "median"])
def test_aggregate_by_duplicates(method):
    # dataframe with duplicates
    full = pd.DataFrame.from_dict(
        {
            "x1": [1.0, 2.0, 3.0, 1.0],
            "x2": [1.0, 2.0, 3.0, 1.0],
            "out1": [4.0, 5.0, 6.0, 3.0],
            "out2": [-4.0, -5.0, -6.0, -3.0],
            "valid_out1": [1, 1, 1, 1],
            "valid_out2": [1, 1, 1, 1],
        }
    )
    expected_aggregated = pd.DataFrame.from_dict(
        {
            "labcode": ["1-4", "2", "3"],
            "x1": [1.0, 2.0, 3.0],
            "x2": [1.0, 2.0, 3.0],
            "out1": [3.5, 5.0, 6.0],
            "out2": [-3.5, -5.0, -6.0],
            "valid_out1": [1, 1, 1],
            "valid_out2": [1, 1, 1],
        }
    )
    domain = Domain(
        inputs=Inputs(features=[if1, if2]), outputs=Outputs(features=[of1, of2])
    )
    aggregated, duplicated_labcodes = domain.aggregate_by_duplicates(
        full, prec=2, method=method
    )
    assert duplicated_labcodes == [["1", "4"]]
    assert_frame_equal(
        aggregated, expected_aggregated, check_dtype=False, check_like=True
    )
    # dataset without duplicates
    full = pd.DataFrame.from_dict(
        {
            "x1": [1.0, 2.0, 3.0, 4.0],
            "x2": [1.0, 2.0, 3.0, 4.0],
            "out1": [4.0, 5.0, 6.0, 3.0],
            "out2": [-4.0, -5.0, -6.0, -3.0],
            "valid_out1": [1, 1, 1, 1],
            "valid_out2": [1, 1, 1, 1],
        }
    )
    expected_aggregated = pd.DataFrame.from_dict(
        {
            "labcode": ["0", "1", "2", "3"],
            "x1": [1.0, 2.0, 3.0, 4.0],
            "x2": [1.0, 2.0, 3.0, 4.0],
            "out1": [4.0, 5.0, 6.0, 3.0],
            "out2": [-4.0, -5.0, -6.0, -3.0],
            "valid_out1": [1, 1, 1, 1],
            "valid_out2": [1, 1, 1, 1],
        }
    )
    domain = Domain(
        inputs=Inputs(features=[if1, if2]), outputs=Outputs(features=[of1, of2])
    )
    aggregated, duplicated_labcodes = domain.aggregate_by_duplicates(
        full, prec=2, method=method
    )
    assert duplicated_labcodes == []


def test_aggregate_by_duplicates_error():
    full = pd.DataFrame.from_dict(
        {
            "x1": [1.0, 2.0, 3.0, 1.0],
            "x2": [1.0, 2.0, 3.0, 1.0],
            "out1": [4.0, 5.0, 6.0, 3.0],
            "out2": [-4.0, -5.0, -6.0, -3.0],
            "valid_out1": [1, 1, 1, 1],
            "valid_out2": [1, 1, 1, 1],
        }
    )
    domain = Domain(
        inputs=Inputs(features=[if1, if2]), outputs=Outputs(features=[of1, of2])
    )
    with pytest.raises(ValueError, match="Unknown aggregation type provided: 25"):
        domain.aggregate_by_duplicates(full, prec=2, method="25")


domain = Domain(
    inputs=Inputs(features=[if1, if2]), outputs=Outputs(features=[of1, of2, of1_, of2_])
)


@pytest.mark.parametrize(
    "domain, FeatureType, exact, expected",
    [
        (domain, Output, True, []),
        (domain, Output, False, [of1, of2, of1_, of2_]),
        (domain, Output, None, [of1, of2, of1_, of2_]),
        (domain, ContinuousOutput, True, [of1, of2, of1_, of2_]),
        (domain, ContinuousOutput, False, [of1, of2, of1_, of2_]),
        (domain, ContinuousOutput, None, [of1, of2, of1_, of2_]),
        (domain, Input, True, []),
        (domain, Input, False, [if1, if2]),
        (domain, Input, None, [if1, if2]),
    ],
)
def test_get_features(domain, FeatureType, exact, expected):
    assert domain.get_features(FeatureType, exact=exact).features == expected


@pytest.mark.parametrize(
    "domain, includes, excludes, exact, expected",
    [
        (domain, [Objective], [], True, []),
        (domain, [Objective], [], False, [of1, of2]),
        (domain, [TargetObjective], [], False, [of1, of2]),
        (domain, [], [Objective], False, [of1_, of2_]),
    ],
)
def test_get_outputs_by_objective(domain: Domain, includes, excludes, exact, expected):
    assert (
        domain.outputs.get_by_objective(
            includes=includes,
            excludes=excludes,
            exact=exact,
        ).features
        == expected
    )


def test_get_outputs_by_objective_none():
    outputs = Outputs(
        features=[
            ContinuousOutput(key="a", objective=None),
            ContinuousOutput(
                key="b", objective=MaximizeSigmoidObjective(w=1, steepness=1, tp=0)
            ),
            ContinuousOutput(key="c", objective=MaximizeObjective()),
        ]
    )
    keys = outputs.get_keys_by_objective(excludes=ConstrainedObjective)
    assert keys == ["c"]
    assert outputs.get_keys().index("c") == 2
    assert outputs.get_keys_by_objective(excludes=Objective, includes=[]) == ["a"]
    assert outputs.get_by_objective(excludes=Objective, includes=[]) == Outputs(
        features=[ContinuousOutput(key="a", objective=None)]
    )


@pytest.mark.parametrize(
    "domain, FeatureType, exact, expected",
    [
        (domain, Output, True, []),
        (domain, Output, False, ["out1", "out2", "out3", "out4"]),
        (domain, Output, None, ["out1", "out2", "out3", "out4"]),
        (domain, ContinuousOutput, True, ["out1", "out2", "out3", "out4"]),
        (domain, ContinuousOutput, None, ["out1", "out2", "out3", "out4"]),
        (domain, Input, True, []),
        (domain, Input, False, ["x1", "x2"]),
        (domain, Input, None, ["x1", "x2"]),
    ],
)
def test_get_feature_keys(domain, FeatureType, exact, expected):
    assert domain.get_feature_keys(FeatureType, exact=exact) == expected


@pytest.mark.parametrize(
    "domain, includes, excludes, exact, expected",
    [
        (domain, [Objective], [], False, ["out1", "out2"]),
        (domain, [TargetObjective], [], False, ["out1", "out2"]),
        (domain, [Objective], [], True, []),
        (domain, [], [Objective], False, ["out3", "out4"]),
    ],
)
def test_get_output_keys_by_objective(
    domain: Domain, includes, excludes, exact, expected
):
    assert (
        domain.outputs.get_keys_by_objective(
            includes=includes,
            excludes=excludes,
            exact=exact,
        )
        == expected
    )


@pytest.mark.parametrize(
    "domain, feature_keys",
    [
        (domain, ["x1", "x2", "out1", "out2"]),
        (domain, ["x1", "x2", "out1"]),
        (domain, ["x1", "x2", "out1", "out3"]),
        (domain, ["x1", "out1", "out3", "out4"]),
        (domain2, ["x1", "x2", "out1", "out2"]),
    ],
)
def test_get_subdomain(domain, feature_keys):
    subdomain = get_subdomain(domain, feature_keys)
    assert subdomain.get_feature_keys(Feature) == feature_keys


@pytest.mark.parametrize(
    "domain, output_feature_keys",
    [
        (domain, []),
        (domain, ["out1", "out2"]),
        (domain, ["x1", "x2"]),
        (domain, ["out1", "f1"]),
        (domain2, ["x1", "out1", "out2"]),
    ],
)
def test_get_subdomain_invalid(domain, output_feature_keys):
    with pytest.raises((AssertionError, ValueError, KeyError)):
        get_subdomain(domain, output_feature_keys)
